###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import sys
import copy
from datetime import datetime

from vistrails.core.data_structures.graph import Graph
from vistrails.core.system import time_strptime
from vistrails.db.versions.v0_5_0.domain import DBVistrail, DBAction, DBTag, DBModule, \
    DBConnection, DBPortSpec, DBFunction, DBParameter, DBLocation, DBAdd, \
    DBChange, DBDelete, DBAnnotation, DBPort

def convertDate(date):
    if date is not None and date != '':
        return datetime(*time_strptime(date, '%d %b %Y %H:%M:%S')[0:6])
    return datetime(1900, 1, 1)

def translateVistrail(_vistrail):
    vistrail = DBVistrail()
    for _action in _vistrail.db_get_actions():
#        print 'translating action %s' % _action.db_time
        functionName = 'translate%s%sAction' % \
            (_action.db_what[0].upper(), _action.db_what[1:])
        thisModule = sys.modules[__name__]
        action = getattr(thisModule, functionName)(_action)
        vistrail.db_add_action(action)
    for _tag in _vistrail.db_get_tags():
        tag = DBTag(time=_tag.db_time,
                    name=_tag.db_name)
        vistrail.db_add_tag(tag)

    convertIds(vistrail)
#     for action in vistrail.getActions():
#         print '%s %s' % (action.id, action.operations)
    vistrail.db_version = '0.5.0'
    return vistrail

def translateAddModuleAction(_action):
    operations = []
    for _module in _action.db_datas:
        module = DBModule(id=_module.db_id,
                          name=_module.db_name,
                          cache=1,
                          location=DBLocation(id=_module.db_id,
                                              x=_module.db_x,
                                              y=_module.db_y))
        module.db_location.relative = False
        operation = DBAdd(id=_action.db_time,
                          what='module',
                          objectId=_module.db_id,
                          data=module)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateAddConnectionAction(_action):
    operations = []
    for _connection in _action.db_datas:
        source = DBPort(id=_connection.db_id,
                        type='source',
                        moduleId=_connection.db_sourceId,
                        moduleName=_connection.db_sourceModule,
                        sig=_connection.db_sourcePort)
        destination = DBPort(id=_connection.db_id,
                             type='destination',
                             moduleId=_connection.db_destinationId,
                             moduleName=_connection.db_destinationModule,
                             sig=_connection.db_destinationPort)
        connection = DBConnection(id=_connection.db_id,
                                  ports=[source, destination])
        operation = DBAdd(id=_action.db_time,
                          what='connection',
                          objectId=_connection.db_id,
                          data=connection)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateChangeParameterAction(_action):
    operations = []
    for _set in _action.db_datas:
        parameter = DBParameter(id=_set.db_parameterId,
                                pos=_set.db_parameterId,
                                name=_set.db_parameter,
                                alias=_set.db_alias,
                                val=_set.db_value,
                                type=_set.db_type)
        function = DBFunction(id=_set.db_functionId,
                              pos=_set.db_functionId,
                              name=_set.db_function,
                              parameters=[parameter])
        operation = DBChange(id=_action.db_time,
                             what='function',
                             oldObjId=_set.db_functionId,
                             parentObjId=_set.db_moduleId,
                             parentObjType='module',
                             data=function)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateAddModulePortAction(_action):
    operations = []
    for _portSpec in _action.db_datas:
        # ids need to be checked
        portSpec = DBPortSpec(id=_portSpec.db_moduleId,
                              name=_portSpec.db_portName,
                              type=_portSpec.db_portType,
                              spec=_portSpec.db_portSpec)
        operation = DBAdd(id=_action.db_time,
                          what='portSpec',
                          objectId=(_portSpec.db_portName,
                                    _portSpec.db_portType),
                          parentObjId=_portSpec.db_moduleId,
                          parentObjType='module',
                          data=portSpec)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateChangeAnnotationAction(_action):
    operations = []
    for _annotation in _action.db_datas:
        if _annotation.db_key.strip() != '' or \
                _annotation.db_value.strip() != '':
            annotation = DBAnnotation(id=-1,
                                      key=_annotation.db_key,
                                      value=_annotation.db_value)
            operation = DBChange(id=_action.db_time,
                                 what='annotation',
                                 oldObjId=_annotation.db_key,
                                 parentObjId=_annotation.db_moduleId,
                                 parentObjType='module',
                                 data=annotation)
            operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateDeleteModuleAction(_action):
    operations = []
    for _module in _action.db_datas:
        operation = DBDelete(id=_action.db_time,
                             what='module',
                             objectId=_module.db_moduleId)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateDeleteConnectionAction(_action):
    operations = []
    for _connection in _action.db_datas:
        operation = DBDelete(id=_action.db_time,
                             what='connection',
                             objectId=_connection.db_connectionId)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateDeleteFunctionAction(_action):
    operations = []
    for _function in _action.db_datas:
        operation = DBDelete(id=_action.db_time,
                             what='function',
                             objectId=_function.db_functionId,
                             parentObjId=_function.db_moduleId,
                             parentObjType='module')
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateDeleteAnnotationAction(_action):
    operations = []
    for _annotation in _action.db_datas:
        operation = DBDelete(id=_action.db_time,
                             what='annotation',
                             objectId=_annotation.db_key)
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateDeleteModulePortAction(_action):
    operations = []
    for _portSpec in _action.db_datas:
        operation = DBDelete(id=_action.db_time,
                             what='portSpec',
                             objectId=(_portSpec.db_portName,
                                       _portSpec.db_portType),
                             parentObjId=_portSpec.db_moduleId,
                             parentObjType='module')
        operations.append(operation)

    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action

def translateMoveModuleAction(_action):
    operations = []
    for _location in _action.db_datas:
        location = DBLocation(id=_location.db_id,
                              x=_location.db_dx,
                              y=_location.db_dy)
        location.relative = True

        operation = DBChange(id=_action.db_time,
                             what='location',
                             oldObjId=_location.db_id,
                             parentObjId=_location.db_id,
                             parentObjType='module',
                             data=location)
        operations.append(operation)
    
    action = DBAction(id=_action.db_time,
                      prevId=_action.db_parent,
                      date=convertDate(_action.db_date),
                      user=_action.db_user,
                      operations=operations)
    return action


### UPDATE IDS ###

def convertIds(vistrail):
    actions = vistrail.db_get_actions()
    actions.sort(key=lambda x: x.db_id)
    objectDict = {}
#    refDict = {'objectDict': objectDict}

    graph = Graph()
    for action in actions:
        graph.add_vertex(action.db_id)
        graph.add_edge(action.db_prevId, action.db_id)

    def convertAction(actionId):
        if actionId == 0:
            return

        allOps = []
        action = vistrail.db_get_action(actionId)
#         objectDict = refDict['objectDict']
#         if action.actionType == 'delete' or action.actionType == 'change':
#             action.objectDict = copy.deepcopy(objectDict)
#         else:
#             action.objectDict = objectDict
        for operation in action.db_get_operations():
            allOps.extend(convertOperation(vistrail,
                                           objectDict,
                                           operation.vtType,
                                           operation))
        action.db_operations = allOps

    def removeObjects(actionId):
        if actionId == 0:
            return
#        print "removeObjects(%s)" % actionId
        action = vistrail.db_get_action(actionId)

        # need to reverse ops here
        reverseOps = action.db_get_operations()
        reverseOps.reverse()
        for operation in reverseOps:
            parentList = getTypeIdList(operation)
            removeObject(operation.db_what,
                         operation.db_oldId,
                         objectDict,
                         parentList[:-1])
        reverseOps.reverse()


    graph.dfs(enter_vertex=convertAction,
              leave_vertex=removeObjects)

def getTypeIdList(operation):
    if operation.db_what in ('module', 'connection'):
        return [(operation.db_what, operation.db_oldId)]
    elif operation.db_what in \
            ('function', 'portSpec', 'location', 'annotation'):
        return [('module', operation.db_oldParentId),
                (operation.db_what, operation.db_oldId)]
    elif operation.db_what in ('port'):
        return [('connection', operation.db_oldParentId),
                (operation.db_what, operation.db_oldId)]
    elif operation.db_what in ('parameter'):
        return [('module', operation.db_moduleId),
                ('function', operation.db_oldParentId),
                ('parameter', operation.db_oldId)]
    else:
        print "unknown type: '%s'" % operation.db_what
        return [(operation.db_what, operation.db_oldId)]

def getOldId(object):
    if object.vtType == 'annotation':
        return object.db_key
    elif object.vtType == 'port':
        return object.db_type
    elif object.vtType == 'portSpec':
        return (object.db_name, object.db_type)
    else:
        return object.getPrimaryKey()

def getChildren(object):
    childList = []
    if object.vtType == 'module':
        childList = object.db_get_functions() + \
            object.db_get_portSpecs() + \
            object.db_get_annotations()
        childList.append(object.db_location)
        object.db_functions = []
        object.db_portSpecs = []
        object.db_annotations = {}
        object.db_location = None
    elif object.vtType == 'connection':
         childList = object.db_get_ports()
         object.db_ports = []
    elif object.vtType == 'function':
        childList =  object.db_get_parameters()
        object.db_parameters = []
    return childList

def captureObject(object, objectDict, newId, parentList):
    
#    print "capturing %s" % object
    currentDict = objectDict
    for key in parentList:
        (objType, objId) = key
#        (currentId, newDict, _) = currentDict[(objType, objId)]
#        currentDict = newDict
        (objList, curIdx) = currentDict[(objType, objId)]
        currentDict = objList[curIdx][1]
    oldId = getOldId(object)
#    print "capture: %s %s" % (object.vtType, oldId)
#    currentDict[(object.vtType, oldId)] = (newId, {}, object)
    if not currentDict.has_key((object.vtType, oldId)):
        currentDict[(object.vtType, oldId)] = ([], -1)
    (curList, curIdx) = currentDict[(object.vtType, oldId)]
    curList.append((newId, {}, object, curIdx))
    currentDict[(object.vtType, oldId)] = (curList, len(curList) - 1)

def captureDelete(objType, objId, objectDict, parentList):
    currentDict = objectDict
    for (aType, aId) in parentList:
#        (currentId, newDict, _) = currentDict[(objType, objId)]
#        currentDict = newDict
        (objList, curIdx) = currentDict[(aType, aId)]
        currentDict = objList[curIdx][1]

#    print "captureDelete: %s %s" % (objType, objId)
    if not currentDict.has_key((objType, objId)):
        raise Exception("invalid delete")
    (curList, curIdx) = currentDict[(objType, objId)]
    curList.append((-1, {}, None, curIdx))
    currentDict[(objType, objId)] = (curList, len(curList) - 1)

def removeObject(oldObjType, oldId, objectDict, parentList):
#     print '%s %s' % (oldObjType, oldId)
#     print objectDict
#     print parentList

    try:
        currentDict = objectDict
        for key in parentList:
            (objType, objId) = key
#            (currentId, newDict, _) = currentDict[(objType, objId)]
#            currentDict = newDict
            (objList, objIdx) = currentDict[(objType, objId)]
            currentDict = objList[objIdx][1]
#        print "remove: %s %s" % (oldObjType, oldId)
        (curList, curIdx) = currentDict[(oldObjType, oldId)]
#        print "ok"
        newIdx = curList[curIdx][3]
#        del curList[curIdx]
        currentDict[(oldObjType, oldId)] = (curList, newIdx)
    except KeyError:
        print "cannot remove (%s, %s)" % (oldObjType, oldId)
        print parentList
        print objList
        print "index: %s"  % objIdx

def findNewId(typeIdList, objectDict):
    try:
        currentDict = objectDict
        for key in typeIdList:
#            (currentId, currentDict, currentObj) = currentDict[key]
            (objList, curIdx) = currentDict[key]
            if curIdx == -1:
                return (None, None)
            (currentId, currentDict, currentObj, _) = objList[curIdx]
        if currentId == -1:
            return (None, None)
        return (currentId, currentObj)
    except KeyError:
        pass
    return (None, None)

def getChildList(typeIdList, objectDict):
    try:
        currentDict = objectDict
        for (objType, objOldId) in typeIdList:
#            (currentId, currentDict, _) = currentDict[(objType, objOldId)]
            (objList, curIdx) = currentDict[(objType, objOldId)]
            if curIdx == -1:
                return {}
            currentDict = objList[curIdx][1]
        return currentDict
    except KeyError:
        pass
    return {}

def createOperation(actionType, objId, objType, parentId, parentType,
                    object=None):
    if actionType == 'add':
        operation = DBAdd(what=objType,
                          objectId=objId,
                          parentObjId=parentId,
                          parentObjType=parentType,
                          data=object)

    elif actionType == 'change':
        operation = DBChange(what=objType,
                             oldObjId=objId,
                             parentObjId=parentId,
                             parentObjType=parentType,
                             data=object)
    elif actionType == 'delete':
        operation = DBDelete(what=objType,
                             objectId=objId,
                             parentObjId=parentId,
                             parentObjType=parentType)
    else:
        msg = "Cannot find actionType='%s'" % actionType
        raise Exception(msg)

    return operation

def convertChangeToAdd(operation):
    return DBAdd(what=operation.db_what,
                 objectId=operation.db_newObjId,
                 parentObjId=operation.db_parentObjId,
                 parentObjType=operation.db_parentObjType,
                 data=operation.db_data)
    
def convertOperation(vistrail, objectDict, actionType, operation):
    newOps = []
    if actionType == 'add':
        object = operation.db_data
        if object.vtType == 'parameter' and object.db_pos == -1:
            return newOps
        operation.db_oldId = operation.db_objectId
        if operation.db_what == 'annotation':
            operation.db_oldId = object.db_key
        elif operation.db_what == 'port':
            operation.db_oldId = object.db_type
        operation.db_oldParentId = operation.db_parentObjId
        parentList = getTypeIdList(operation)


        newId = vistrail.idScope.getNewId(object.vtType)
        captureObject(object, objectDict, newId, parentList[:-1])
        operation.db_objectId = newId
        oldId = object.getPrimaryKey()
        if object.vtType == 'annotation':
            oldId = object.db_key
        elif object.vtType == 'port':
            oldId = object.db_type
        if hasattr(object, 'db_id'):
            object.db_id = newId

        # set parent ids correctly...
        operation.db_id = vistrail.idScope.getNewId('operation')
        if operation.db_parentObjId is not None:
            oldParentObjId = operation.db_parentObjId
            operation.db_parentObjId = findNewId(parentList[:-1], objectDict)[0]
        if object.vtType == 'port':
            object.db_moduleId = \
                findNewId([('module', object.db_moduleId)], objectDict)[0]
#         if object.vtType == 'connection':
#             for port in object.db_ports.itervalues():
#                 port.db_moduleId = \
#                     findNewId([('module', port.db_moduleId)], objectDict)[0]

        newOps.append(operation)

        # set child operations
        children = getChildren(object)
        for child in children:
            # hack to get around fact that location ids are wrong
            if child.vtType == 'location':
                child.db_id = oldId
            newOp = createOperation('add',
                                    child.getPrimaryKey(),
                                    child.vtType,
                                    oldId,
                                    object.vtType,
                                    child)

            # hack to get moduleId at parameter level
            if child.vtType == 'parameter':
                newOp.db_moduleId = oldParentObjId
            newOps.extend(convertOperation(vistrail, 
                                           objectDict, 
                                           'add',
                                           newOp))
            newOp.db_parentObjId = newId
    elif actionType == 'change':
        object = operation.db_data
        if object.vtType == 'parameter' and object.db_pos == -1:
            return newOps
        operation.db_oldId = operation.db_oldObjId
        if operation.db_what == 'annotation':
            operation.db_oldId = object.db_key
        elif operation.db_what == 'port':
            operation.db_oldId = object.db_type
        operation.db_oldParentId = operation.db_parentObjId
        parentList = getTypeIdList(operation)

        # need to get changed id as new id if have one
        (foundId, foundObj) = findNewId(parentList, objectDict)
        if foundId is not None:
            if foundObj.vtType == 'function' and \
                    foundObj.db_pos == object.db_pos and \
                    foundObj.db_name == object.db_name:
                # don't create new function, convert parameter
                for parameter in object.db_parameters:
                    newOp = createOperation('change',
                                            parameter.getPrimaryKey(),
                                            parameter.vtType,
                                            object.getPrimaryKey(),
                                            object.vtType,
                                            parameter)
                    newOp.db_moduleId = operation.db_parentObjId
                    newOps.extend(convertOperation(vistrail,
                                                   objectDict,
                                                   'change',
                                                   newOp))
                    newOp.db_parentObjId = foundId
                return newOps
            else:
                if foundObj.vtType == 'location' and object.relative == True:
                    object.db_x += foundObj.db_x
                    object.db_y += foundObj.db_y
                    object.relative = False
                # get new id for new object
                newId = vistrail.idScope.getNewId(object.vtType)
                operation.db_oldObjId = foundId
                operation.db_newObjId = newId
        else:
            # get new id for new object
            newId = vistrail.idScope.getNewId(object.vtType)
            operation.db_oldObjId = -1
            operation.db_newObjId = newId
            anOldId = operation.db_oldId
            anOldParentId = operation.db_parentObjId
            if hasattr(operation,'db_moduleId'):
                aModuleId = operation.db_moduleId
            else:
                aModuleId = None
            operation = convertChangeToAdd(operation)
            operation.db_oldId = anOldId
            operation.db_oldParentId = operation.db_parentObjId
            operation.db_moduleId = aModuleId

        # need to do child deletes first
        childDict = getChildList(parentList, objectDict)
        for k,v in childDict.items():
            (objType, objId) = k
#            (newId, newDict) = v
#            print 'creating delete for %s'  % objType
            newOp = createOperation('delete',
                                    objId,
                                    objType,
                                    object.getPrimaryKey(),
                                    object.vtType)
            # hack to get moduleId at parameter level
            if objType == 'parameter':
                newOp.db_moduleId = operation.db_parentObjId
            newOps.extend(convertOperation(vistrail,
                                           objectDict,
                                           'delete',
                                           newOp))
            newOp.db_parentObjId = newId
        # don't reverse -- ordering is correct
        # newOps.reverse()

        # set new object id
        captureObject(object, objectDict, newId, parentList[:-1])
#        operation.db_objectId = newId
        oldId = object.getPrimaryKey()
        if object.vtType == 'annotation':
            oldId = object.db_key
        elif object.vtType == 'port':
            oldId = object.db_type
        if hasattr(object, 'db_id'):
            object.db_id = newId

        # set parent ids correctly...
        operation.db_id = vistrail.idScope.getNewId('operation')
        if operation.db_parentObjId is not None:
            oldParentObjId = operation.db_parentObjId
            operation.db_parentObjId = findNewId(parentList[:-1], objectDict)[0]
        if object.vtType == 'port':
            object.db_moduleId = \
                findNewId([('module', object.db_moduleId)], objectDict)[0]
#         if object.vtType == 'connection':
#             for port in object.db_ports.itervalues():
#                 port.db_moduleId = \
#                     findNewId([('module', port.db_moduleId)], objectDict)[0]
        newOps.append(operation)

        # set child operations
        children = getChildren(operation.db_data)
        for child in children:
#            print 'creating add for %s' % child.vtType
            newOp = createOperation('add',
                                    child.getPrimaryKey(),
                                    child.vtType,
                                    oldId,
                                    object.vtType,
                                    child)
            # hack to get moduleId at parameter level
            if child.vtType == 'parameter':
                newOp.db_moduleId = oldParentObjId
            newOps.extend(convertOperation(vistrail, 
                                           objectDict, 
                                           'add',
                                           newOp))
            newOp.db_parentObjId = newId
    elif actionType == 'delete':
        operation.db_oldId = operation.db_objectId
#         if operation.db_what == 'annotation':
#             operation.db_oldId = object.db_key
#         elif operation.db_what == 'port':
#             operation.db_oldId = object.db_type
        operation.db_oldParentId = operation.db_parentObjId
        parentList = getTypeIdList(operation)

        # get new id for delete operation
        (newId, _) = findNewId(parentList, objectDict)
#        print 'found new id:  %s' % newId
        if newId is None:
            msg = "Cannot find id: %s" % parentList
            print msg
#            raise Exception(msg)
            return []

        # need to do child deletes first
        childDict = getChildList(parentList, objectDict)
        for k,v in childDict.items():
            (objType, objId) = k
#            (newId, newDict) = v
            newOp = createOperation('delete',
                                    objId,
                                    objType,
                                    operation.db_objectId,
                                    operation.db_what)
            # hack to get moduleId at parameter level
            if objType == 'parameter':
                newOp.db_moduleId = operation.db_parentObjId
            newOps.extend(convertOperation(vistrail,
                                           objectDict,
                                           'delete',
                                           newOp))
            newOp.db_parentObjId = newId
#        newOps.reverse()

        captureDelete(operation.db_what, operation.db_objectId, objectDict, 
                      parentList[:-1])
        operation.db_objectId = newId
        
        # set parent ids correctly
        operation.db_id = vistrail.idScope.getNewId('operation')
        if operation.db_parentObjId is not None:
            operation.db_parentObjId = findNewId(parentList[:-1], objectDict)[0]
        newOps.append(operation)

    return newOps



#!/usr/bin/env python

# This is a simple volume rendering example that uses a
# vtkVolumeRayCast mapper

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Create the standard renderer, render window and interactor
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Create the reader for the data
reader = vtk.vtkStructuredPointsReader()
reader.SetFileName(VTK_DATA_ROOT + "/Data/ironProt.vtk")

# Create transfer mapping scalar value to opacity
opacityTransferFunction = vtk.vtkPiecewiseFunction()
opacityTransferFunction.AddPoint(20, 0.0)
opacityTransferFunction.AddPoint(255, 0.2)

# Create transfer mapping scalar value to color
colorTransferFunction = vtk.vtkColorTransferFunction()
colorTransferFunction.AddRGBPoint(0.0, 0.0, 0.0, 0.0)
colorTransferFunction.AddRGBPoint(64.0, 1.0, 0.0, 0.0)
colorTransferFunction.AddRGBPoint(128.0, 0.0, 0.0, 1.0)
colorTransferFunction.AddRGBPoint(192.0, 0.0, 1.0, 0.0)
colorTransferFunction.AddRGBPoint(255.0, 0.0, 0.2, 0.0)

# The property describes how the data will look
volumeProperty = vtk.vtkVolumeProperty()
volumeProperty.SetColor(colorTransferFunction)
volumeProperty.SetScalarOpacity(opacityTransferFunction)
volumeProperty.ShadeOn()
volumeProperty.SetInterpolationTypeToLinear()

# The mapper / ray cast function know how to render the data
compositeFunction = vtk.vtkVolumeRayCastCompositeFunction()
volumeMapper = vtk.vtkVolumeRayCastMapper()
volumeMapper.SetVolumeRayCastFunction(compositeFunction)
volumeMapper.SetInputConnection(reader.GetOutputPort())

# The volume holds the mapper and the property and
# can be used to position/orient the volume
volume = vtk.vtkVolume()
volume.SetMapper(volumeMapper)
volume.SetProperty(volumeProperty)

ren.AddVolume(volume)
ren.SetBackground(1, 1, 1)
renWin.SetSize(600, 600)
renWin.Render()

def CheckAbort(obj, event):
    if obj.GetEventPending() != 0:
        obj.SetAbortRender(1)
 
renWin.AddObserver("AbortCheckEvent", CheckAbort)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import itertools

from vistrails.core.modules.vistrails_module import Module, ModuleError

#################################################################################
## Products

class ElementwiseProduct(Module):
    """This module does the product of two lists.

    If NumericalProduct is True, this will effectively compute the product of
    each element, eg:
        [1, 2, 3] x [3, -3, 1] = [3, -6, 3]
    Else, it will make tuples, eg:
        [1, 2, 3] x [3, -3, 1] = [(1, 3), (2, -3), (3, 1)]
    """

    def compute(self):
        list1 = self.get_input('List1')
        list2 = self.get_input('List2')
        if len(list1) != len(list2):
            raise ModuleError(self, "Both lists must have the same size.")

        numerical = self.get_input('NumericalProduct')
        if numerical:
            result = [a*b for a, b in itertools.izip(list1, list2)]
        else:
            result = zip(list1, list2)

        self.set_output('Result', result)


class Dot(Module):
    """This module produces a Dot product between two lists."""

    def compute(self):
        list1 = self.get_input("List1")
        list2 = self.get_input("List2")
        if len(list1) != len(list2):
            raise ModuleError(self, 'Both lists must have the same size.')

        result = sum(a*b for a, b in itertools.izip(list1, list2))

        self.set_output("Result", result)


class Cross(Module):
    """This module produces a Cross product between two 3-D vectors."""

    def compute(self):
        list1 = self.get_input("List1")
        list2 = self.get_input("List2")
        if not (len(list1) == len(list2) == 3):
            raise ModuleError(self, 'Both lists must have size 3.')

        x1, y1, z1 = list1
        x2, y2, z2 = list2

        result = [y1*z2 - y2*z1,
                  z1*x2 - z2*x1,
                  x1*y2 - x2*y1]

        self.set_output("Result", result)


class CartesianProduct(Module):
    """This module does the cartesian product of two lists.
    """

    def compute(self):
        list1 = self.get_input("List1")
        list2 = self.get_input("List2")
        result = []
        # If CombineTuple is not set or True, existing tuples will be
        # concatenated instead of put inside a new tuple, eg:
        #   with CombineTuple (default):
        #     [(1, 2)], [3, 4] -> [(1, 2, 3), (1, 2, 4)]
        #   without:
        #     [(1, 2)], [3, 4] -> [((1, 2), 3), ((1, 2), 4)]
        if not self.get_input('CombineTuple'):
            for i in list1:
                for j in list2:
                    tuple_ = (i, j)
                    result.append(tuple_)
        else:
            for i in list1:
                for j in list2:
                    if isinstance(i, tuple) and isinstance(j, tuple):
                        tuple_ = i + j
                        result.append(tuple_)
                    elif isinstance(i, tuple) and not isinstance(j, tuple):
                        tuple_ = i + (j,)
                        result.append(tuple_)
                    elif not isinstance(i, tuple) and isinstance(j, tuple):
                        tuple_ = (i,) + j
                        result.append(tuple_)
                    else:
                        tuple_ = (i, j)
                        result.append(tuple_)

        self.set_output("Result", result)


###############################################################################

import unittest

from vistrails.tests.utils import intercept_result, execute


class TestProducts(unittest.TestCase):
    def test_elementwise(self):
        with intercept_result(ElementwiseProduct, 'Result') as results:
            self.assertFalse(execute([
                    ('ElementwiseProduct', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[1, 2, 0]")]),
                        ('List2', [('List', "[4, -3, 7]")]),
                    ]),
                ]))
        self.assertEqual(results, [[4, -6, 0]])

    def test_elementwise_tuples(self):
        with intercept_result(ElementwiseProduct, 'Result') as results:
            self.assertFalse(execute([
                    ('ElementwiseProduct', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[1, 2, 0]")]),
                        ('List2', [('List', "[4, -3, 7]")]),
                        ('NumericalProduct', [('Boolean', "False")]),
                    ]),
                ]))
        self.assertEqual(results, [[(1, 4), (2, -3), (0, 7)]])

    def test_dot(self):
        with intercept_result(Dot, 'Result') as results:
            self.assertFalse(execute([
                    ('Dot', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[1, 2, 0]")]),
                        ('List2', [('List', "[4, -3, 7]")]),
                    ]),
                ]))
        self.assertEqual(results, [-2])

    def test_cross(self):
        with intercept_result(Cross, 'Result') as results:
            self.assertFalse(execute([
                    ('Cross', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[1, 2, 0]")]),
                        ('List2', [('List', "[4, -3, 7]")]),
                    ]),
                ]))
        # ( | 2 -3 |, | 0  7 |, | 1  4 | )
        #   | 0  7 |  | 1  4 |  | 2 -3 |
        self.assertEqual(results, [[14, -7, -11]])

    def test_cartesian_combine(self):
        with intercept_result(CartesianProduct, 'Result') as results:
            self.assertFalse(execute([
                    ('CartesianProduct', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[(1, 2)]")]),
                        ('List2', [('List', "[3, 4]")]),
                    ]),
                ]))
        self.assertEqual(results, [[(1, 2, 3), (1, 2, 4)]])

    def test_cartesian_nocombine(self):
        with intercept_result(CartesianProduct, 'Result') as results:
            self.assertFalse(execute([
                    ('CartesianProduct', 'org.vistrails.vistrails.control_flow', [
                        ('List1', [('List', "[(1, 2)]")]),
                        ('List2', [('List', "[3, 4]")]),
                        ('CombineTuple', [('Boolean', "False")]),
                    ]),
                ]))
        self.assertEqual(results, [[((1, 2), 3), ((1, 2), 4)]])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.core.data_structures.point import Point

import unittest
import random

################################################################################
# Rect

class Rect(object):
    def __init__(self, lower_left=None, upper_right=None):
        """ Rect(lower_left: Point, upper_right: Point) -> Rect
        Creates a Rect given the lower_left and the upper_right points. It creates
        a copy of the given points. Return a Rect

        """        
        if lower_left is None:
            self.lower_left = Point()
        else:
            self.lower_left = copy.copy(lower_left)

        if upper_right is None:
            self.upper_right = Point()
        else:
            self.upper_right = copy.copy(upper_right)
        self.check_extent()

    @staticmethod
    def create(left, right, down, up):
        """ create(left: float, right: float, down: float, up: float) -> Point
        Creates a Rect from four float extents and return a Rect

        """
        return Rect(Point(min(left, right),
                          min(up, down)),
                    Point(max(left, right),
                          max(down, up)))
    
    def set_left(self, x):
        """ set_left(x: float) -> None
        Sets the left limit of the Rect and return nothing

        """
        self.lower_left.x = x

    def set_right(self, x):
        """ set_right(x: float) -> None
        Sets the right limit of the Rect and return nothing

        """
        self.upper_right.x = x

    def set_up(self,y):
        """ set_up(y: float) -> None
        Sets the upper limit of the Rect and return nothing

        """
        self.upper_right.y = y

    def set_down(self, y):
        """ set_down(y: float) -> None
        Sets the lower limit of the Rect and return nothing

        """
        self.lower_left.y = y

    def center(self):
        """ center() -> Point
        Compute the center of the Rect and return a Point

        """
        return (self.upper_right + self.lower_left) * 0.5

    def check_extent(self):
        """ check_extent() -> None
        Makes sure left limit is less than right limit, and lower limit is less
        than upper limit. In other words, ensures that, immediately after the
        call:
        self.lower_left.x <= self.upper_right.x and
        self.lower_left.y <= self.upper_right.y

        """
        if self.lower_left.x > self.upper_right.x:
            dlx = self.lower_left.x
            self.lower_left.x = self.upper_right.x
            self.upper_right.y = dlx

        if self.lower_left.y > self.upper_right.y:
            dly = self.lower_left.y
            self.lower_left.y = self.upper_right.y
            self.upper_right.y = dly

################################################################################
# Unit tests


class TestRect(unittest.TestCase):

    def test_create(self):
        """Exercises Rect.create()"""
        for i in xrange(100):
            a = random.uniform(-1.0, 1.0)
            b = random.uniform(-1.0, 1.0)
            c = random.uniform(-1.0, 1.0)
            d = random.uniform(-1.0, 1.0)

            r = Rect.create(a, b, c, d)
            assert r.lower_left.x == min(a, b)
            assert r.upper_right.x == max(a, b)
            assert r.lower_left.y == min(c, d)
            assert r.upper_right.y == max(c, d)

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import os
import sys
import getpass
import socket
import datetime
import platform
import tempfile
import warnings

from vistrails.core import debug
from vistrails.core.utils import VistrailsDeprecation


__all__ = ['touch', 'mkdir', 'python_version',
           'current_user', 'current_ip', 'current_time',
           'current_machine', 'current_architecture', 'current_processor',
           'get_elementtree_library', 'temporary_directory']

###############################################################################

def touch(file_name):
    """touch(file_name) -> None Equivalent to 'touch' in a shell. If
    file exists, updates modified time to current time. If not,
    creates a new 0-length file.

    """
    if os.path.isfile(file_name):
        os.utime(file_name, None)
    else:
        open(file_name, 'w')

def mkdir(dir_name):
    """mkdir(dir_name) -> None Equivalent to 'mkdir' in a shell except
    that if the directory exists, it will not be overwritten.

    """
    if not os.path.isdir(dir_name):
        os.mkdir(dir_name)

def python_version():
    """python_version() -> (major, minor, micro, release, serial)
    Returns python version info."""
    return sys.version_info

def current_user():
    return getpass.getuser()

def current_ip():
    """ current_ip() -> str
    Gets current IP address trying to avoid the IPv6 interface """
    try:
        info = socket.getaddrinfo(socket.gethostname(), None)
        # Try to find an IPv4
        for i in info:
            if i[0] == socket.AF_INET:
                return i[4][0]
        # Return any address
        for i in info:
            if i[0] in (socket.AF_INET, socket.AF_INET6):
                return i[4][0]
    except Exception, e:
        debug.unexpected_exception(e)
        return ''

def current_time():
    """current_time() -> datetime.datetime
    Returns the current time

    """
    # FIXME should use DB if available...
    return datetime.datetime.now()

def current_machine():
    return socket.getfqdn()

def current_architecture():
    bit_string = platform.architecture()[0]
    if bit_string.endswith('bit'):
        return int(bit_string[:-3])
    else:
        return 32 # default value

def current_processor():
    proc = platform.processor()
    if not proc:
        proc = 'n/a'
    return proc

def get_elementtree_library():
    try:
        import cElementTree as ElementTree
    except ImportError:
        # try python 2.5-style
        import xml.etree.cElementTree as ElementTree
    return ElementTree

def temporary_directory():
    warnings.warn(
            "temporary_directory() is deprecated; use the tempfile module "
            "instead",
            category=VistrailsDeprecation)
    return tempfile.gettempdir()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from auto_gen import DBWorkflow as _DBWorkflow
from id_scope import IdScope

import copy

class DBWorkflow(_DBWorkflow):

    def __init__(self, *args, **kwargs):
        _DBWorkflow.__init__(self, *args, **kwargs)
        self.objects = {}
        self.tmp_id = IdScope(1)

    def __copy__(self):
        return DBWorkflow.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = _DBWorkflow.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = DBWorkflow
        # need to go through and reset the index to the copied objects
        cp.objects = {}
        for (child, _, _) in cp.db_children():
            cp.addToIndex(child)
        cp.tmp_id = copy.copy(self.tmp_id)
        return cp        

    def addToIndex(self, object):
        self.objects[(object.vtType, object.getPrimaryKey())] = object

    def deleteFromIndex(self, object):
        del self.objects[(object.vtType, object.getPrimaryKey())]

    def capitalizeOne(self, str):
        return str[0].upper() + str[1:]

    def db_print_objects(self):
        for k,v in self.objects.iteritems():
            print '%s: %s' % (k, v)

    def db_get_object(self, type, id):
        return self.objects[(type, id)]

    def db_add_object(self, object, parentObjType=None, parentObjId=None):
        if parentObjType is None or parentObjId is None:
            parentObj = self
        else:
            try:
                parentObj = self.objects[(parentObjType, parentObjId)]
            except KeyError:
                msg = "Cannot find object of type '%s' with id '%s'" % \
                    (parentObjType, parentObjId)
                raise Exception(msg)
        funname = 'db_add_%s' % object.vtType
        objCopy = copy.copy(object)
        getattr(parentObj, funname)(objCopy)
        self.addToIndex(objCopy)

    def db_change_object(self, object, parentObjType=None, parentObjId=None):
        if parentObjType is None or parentObjId is None:
            parentObj = self
        else:
            try:
                parentObj = self.objects[(parentObjType, parentObjId)]
            except KeyError:
                msg = "Cannot find object of type '%s' with id '%s'" % \
                    (parentObjType, parentObjId)
                raise Exception(msg)
        funname = 'db_change_%s' % object.vtType
        objCopy = copy.copy(object)
        getattr(parentObj, funname)(objCopy)
        self.addToIndex(objCopy)

    def db_delete_object(self, objId, objType, 
                         parentObjType=None, parentObjId=None):
        if parentObjType is None or parentObjId is None:
            parentObj = self
        else:
            try:
                parentObj = self.objects[(parentObjType, parentObjId)]
            except KeyError:
                msg = "Cannot find object of type '%s' with id '%s'" % \
                    (parentObjType, parentObjId)
                raise Exception(msg)
        funname = 'db_get_%s' % objType
        try:
            object = getattr(parentObj, funname)(objId)
        except AttributeError:
            attr_name = 'db_%s' % objType
            object = getattr(parentObj, attr_name)
        funname = 'db_delete_%s' % objType
        getattr(parentObj, funname)(object)
        self.deleteFromIndex(object)



#!/usr/bin/env python

# Perform psuedo volume rendering in a structured grid by compositing
# translucent cut planes. This same trick can be used for unstructured
# grids. Note that for better results, more planes can be created. Also,
# if your data is vtkImageData, there are much faster methods for volume
# rendering.

import vtk
from vtk.util.colors import *
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Create pipeline. Read structured grid data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# A convenience, use this filter to limit data for experimentation.
extract = vtk.vtkExtractGrid()
extract.SetVOI(1, 55, -1000, 1000, -1000, 1000)
extract.SetInputConnection(pl3d.GetOutputPort())

# The (implicit) plane is used to do the cutting
plane = vtk.vtkPlane()
plane.SetOrigin(0, 4, 2)
plane.SetNormal(0, 1, 0)

# The cutter is set up to process each contour value over all cells
# (SetSortByToSortByCell). This results in an ordered output of polygons
# which is key to the compositing.
cutter = vtk.vtkCutter()
cutter.SetInputConnection(extract.GetOutputPort())
cutter.SetCutFunction(plane)
cutter.GenerateCutScalarsOff()
cutter.SetSortByToSortByCell()

clut = vtk.vtkLookupTable()
clut.SetHueRange(0, .67)
clut.Build()

cutterMapper = vtk.vtkPolyDataMapper()
cutterMapper.SetInputConnection(cutter.GetOutputPort())
cutterMapper.SetScalarRange(.18, .7)
cutterMapper.SetLookupTable(clut)

cut = vtk.vtkActor()
cut.SetMapper(cutterMapper)

# Add in some surface geometry for interest.
iso = vtk.vtkContourFilter()
iso.SetInputConnection(pl3d.GetOutputPort())
iso.SetValue(0, .22)
normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(iso.GetOutputPort())
normals.SetFeatureAngle(45)
isoMapper = vtk.vtkPolyDataMapper()
isoMapper.SetInputConnection(normals.GetOutputPort())
isoMapper.ScalarVisibilityOff()
isoActor = vtk.vtkActor()
isoActor.SetMapper(isoMapper)
isoActor.GetProperty().SetDiffuseColor(tomato)
isoActor.GetProperty().SetSpecularColor(white)
isoActor.GetProperty().SetDiffuse(.8)
isoActor.GetProperty().SetSpecular(.5)
isoActor.GetProperty().SetSpecularPower(30)

outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineTubes = vtk.vtkTubeFilter()
outlineTubes.SetInputConnection(outline.GetOutputPort())
outlineTubes.SetRadius(.1)

outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outlineTubes.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Create the RenderWindow, Renderer and Interactor
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
outlineActor.GetProperty().SetColor(banana)
ren.AddActor(isoActor)
isoActor.VisibilityOn()
ren.AddActor(cut)
opacity = .1
cut.GetProperty().SetOpacity(1)
ren.SetBackground(1, 1, 1)
renWin.SetSize(640, 480)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 50)
cam1.SetFocalPoint(9.71821, 0.458166, 29.3999)
cam1.SetPosition(2.7439, -37.3196, 38.7167)
cam1.ComputeViewPlaneNormal()
cam1.SetViewUp(-0.16123, 0.264271, 0.950876)

# Cut: generates n cut planes normal to camera's view plane
def Cut(n):
    global cam1, opacity
    plane.SetNormal(cam1.GetViewPlaneNormal())
    plane.SetOrigin(cam1.GetFocalPoint())
    cutter.GenerateValues(n, -5, 5)
    clut.SetAlphaRange(opacity, opacity)
    renWin.Render()
 

# Generate 10 cut planes
Cut(20)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import base64
import getpass
import os.path
from vistrails.core import get_vistrails_application
from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core.system import vistrails_default_file_type, get_elementtree_library, \
                        default_connections_file, vistrails_examples_directory
from vistrails.core.external_connection import ExtConnectionList, DBConnection
from vistrails.core.thumbnails import ThumbnailCache
from vistrails.core import debug
from vistrails.db.services.locator import XMLFileLocator as _XMLFileLocator, \
    DBLocator as _DBLocator, ZIPFileLocator as _ZIPFileLocator, \
    BaseLocator as _BaseLocator, UntitledLocator as _UntitledLocator
from vistrails.db.services.io import SaveBundle, test_db_connection
from vistrails.db import VistrailsDBException
from vistrails.db.domain import DBWorkflow
ElementTree = get_elementtree_library()

class BaseLocator(_BaseLocator):
    @staticmethod
    def convert_locator(locator):
        if locator.__class__ == _XMLFileLocator:
            locator.__class__ = XMLFileLocator
        elif locator.__class__ == _ZIPFileLocator:
            locator.__class__ = ZIPFileLocator
        elif locator.__class__ == _DBLocator:
            DBLocator.convert(locator)
        elif locator.__class__ == _UntitledLocator:
            locator.__class__ = UntitledLocator
            
    @staticmethod
    def from_url(url):
        locator = _BaseLocator.from_url(url)
        BaseLocator.convert_locator(locator)
        return locator

class CoreLocator(object):
    @staticmethod
    def prompt_autosave(parent_widget):
        pass # Opens a dialog that prompts the user if they want to
             # use temporaries


    @staticmethod
    def load_from_gui(parent_widget, obj_type):
        pass # Opens a dialog that the user will be able to use to
             # show the right values, and returns a locator suitable
             # for loading a file

    @staticmethod
    def save_from_gui(parent_widget, obj_type, locator):
        pass # Opens a dialog that the user will be able to use to
             # show the right values, and returns a locator suitable
             # for saving a file

    def update_from_gui(self, klass=None):
        pass

    # FIXME Need to do some more intelligent conversions anywhere this
    # function gets called
    @staticmethod
    def get_convert_klass(vt_type):
        from vistrails.core.vistrail.vistrail import Vistrail
        from vistrails.core.vistrail.pipeline import Pipeline
        from vistrails.core.log.log import Log
        from vistrails.core.modules.module_registry import ModuleRegistry
        from vistrails.core.log.opm_graph import OpmGraph
        
        klass_map = {Vistrail.vtType: Vistrail,
                     Pipeline.vtType: Pipeline,
                     Log.vtType: Log,
                     ModuleRegistry.vtType: ModuleRegistry,
                     OpmGraph.vtType: OpmGraph}
        return klass_map[vt_type]

class UntitledLocator(_UntitledLocator, CoreLocator):
    def load(self, klass=None):
        from vistrails.core.vistrail.vistrail import Vistrail
        if klass is None:
            klass = Vistrail
        obj = _UntitledLocator.load(self, klass.vtType)
        klass.convert(obj)
        obj.locator = self
        return obj

class XMLFileLocator(_XMLFileLocator, CoreLocator):

    def __init__(self, filename, **kwargs):
        _XMLFileLocator.__init__(self, filename, **kwargs)
        
    def load(self, klass=None):
        from vistrails.core.vistrail.vistrail import Vistrail
        if klass is None:
            klass = Vistrail
        obj = _XMLFileLocator.load(self, klass.vtType)
        klass.convert(obj)
        obj.locator = self
        return obj

    def save(self, obj):
        is_bundle = False
        if type(obj) == type(SaveBundle(None)):
            is_bundle = True
            save_bundle = obj
            obj = save_bundle.get_primary_obj()
        klass = obj.__class__
        obj = _XMLFileLocator.save(self, obj, False)
        klass.convert(obj)
        obj.locator = self
        if is_bundle:
            return SaveBundle(save_bundle.bundle_type, obj)
        return obj

    def save_as(self, obj, version=None):
        is_bundle = False
        if type(obj) == type(SaveBundle(None)):
            is_bundle = True
            save_bundle = obj
            obj = save_bundle.get_primary_obj()
        klass = obj.__class__
        obj = _XMLFileLocator.save(self, obj, True, version)
        klass.convert(obj)
        obj.locator = self
        if is_bundle:
            return SaveBundle(save_bundle.bundle_type, obj)
        return obj

    ##########################################################################

    def __eq__(self, other):
        if not isinstance(other, XMLFileLocator):
            return False
        return self._name == other._name

    ##########################################################################

    @staticmethod
    def prompt_autosave(parent_widget):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_autosave_prompt(parent_widget)
        
    @staticmethod
    def load_from_gui(parent_widget, obj_type):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_load_file_locator_from_gui(parent_widget, obj_type)

    @staticmethod
    def save_from_gui(parent_widget, obj_type, locator=None):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_save_file_locator_from_gui(parent_widget, obj_type,
                                                         locator)

#    def update_from_gui(self, parent_widget, klass=None):
#        from core.vistrail.vistrail import Vistrail
#        if klass is None:
#            klass = Vistrail
#        import gui.extras.core.db.locator as db_gui
#        return db_gui.get_load_file_locator_from_gui(parent_widget, klass.vtType)

class DBLocator(_DBLocator, CoreLocator):
    class getKeyChain(object):
        def set_key(self, key, passwd):
            get_vistrails_application().keyChain.set_key(key,passwd)
        
        def get_key(self, key):
            return get_vistrails_application().keyChain.get_key(key)
    
    keyChain = getKeyChain()
    
    def __init__(self, host, port, database, user, passwd, name=None,
                 **kwargs):
        _DBLocator.__init__(self, host, port, database, user, passwd, name,
                            **kwargs)
        self.__list = ExtConnectionList.getInstance(default_connections_file())
        self.ext_connection_id = -1

    def load(self, klass=None):
        from vistrails.core.vistrail.vistrail import Vistrail
        if klass is None:
            klass = Vistrail
        save_bundle = _DBLocator.load(self, klass.vtType, ThumbnailCache.getInstance().get_directory())
        if klass.vtType == DBWorkflow.vtType:
            wf = save_bundle
            klass = self.get_convert_klass(wf.vtType)
            klass.convert(wf)
            wf.locator = self
            return wf
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        return save_bundle

    def save(self, save_bundle):
        save_bundle = _DBLocator.save(self, save_bundle, False)
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        return save_bundle

    def save_as(self, save_bundle, version=None):
        save_bundle = _DBLocator.save(self, save_bundle, True, version)
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        # Need to copy images into thumbnail cache directory so references
        # won't become invalid if they are in a temp dir that gets destroyed
        # when the previous locator is closed
        import shutil
        thumb_cache = ThumbnailCache.getInstance()
        thumb_cache_dir = thumb_cache.get_directory()
        new_thumbnails = []
        for thumbnail in save_bundle.thumbnails:
            if os.path.dirname(thumbnail) == thumb_cache_dir:
                new_thumbnails.append(thumbnail)
            else:
                cachedir_thumbnail = os.path.join(thumb_cache_dir, os.path.basename(thumbnail))
                try:
                    shutil.copyfile(thumbnail, cachedir_thumbnail)
                    new_thumbnails.append(cachedir_thumbnail)
                except Exception, e:
                    debug.critical("copying %s -> %s failed" % (
                                   thumbnail, cachedir_thumbnail),
                                   e)
        save_bundle.thumbnails = new_thumbnails
        # Need to update thumbnail cache in case some references have changed
        thumb_cache.add_entries_from_files(save_bundle.thumbnails)
        return save_bundle

    def update_from_gui(self, parent_widget, klass=None):
        from vistrails.core.vistrail.vistrail import Vistrail
        import vistrails.gui.extras.core.db.locator as db_gui
        if klass is None:
            klass = Vistrail
        config = self.find_connection_info(self._host, self._port, self._db) 
        if config is None or config['succeeded']==False:
            config = db_gui.get_db_connection_from_gui(parent_widget,
                                                       -1,
                                                       "",
                                                       self._host,
                                                       self._port,
                                                       self._user,
                                                       self._passwd,
                                                       self._db)
        
        if config is not None and config['succeeded'] == True:
            self._host = config['host']
            self._port = config['port']
            self._db = config['db']
            self._user = config['user']
            self._passwd = config['passwd']
            self.ext_connection_id = self.set_connection_info(**config)
            return True
        
        return False
    
    def update_from_console(self):
        config = self.find_connection_info(self._host, self._port, self._db)
        
        if config is None:
            # the problem here is if VisTrails is being run through command
            # line from LaTex, stdout is being redirected to a log file, so
            # the user does not see the prompt in raw_input. getpass uses the 
            # controlling terminal so it works fine. Just to make sure he sees 
            # the first message prompt we will the controlling terminal
            try:
                f= open('/dev/tty', 'w')
                f.write("\nConnect to db with username [%s]: "%self._user)
                f.close()
                user = raw_input()
            except IOError:
                debug.warning("Couldn't write to terminal. Will try stdout")
                user = raw_input("Connecting to db with username[%s]: "%self._user)
            try:
                if user != '':
                    self._user = user
                passwd = getpass.getpass("password:")
                self._passwd = passwd
                config = {'host': self._host,
                          'port': int(self._port),
                          'user': self._user,
                          'passwd': self._passwd,
                          'db': self._db
                          }
                test_db_connection(config)
                config['succeeded'] = True
                config['name'] = '%s@%s'%(self._user,self._host)
                config['id'] = -1
            except VistrailsDBException, e:
                debug.critical('VisTrails DB Exception',  e)
                config['succeeded'] = False
            except Exception, e2:
                debug.critical('VisTrails Exception', e2)
                config['succeeded'] = False
        if config is not None:
            if config['succeeded'] == False:
                passwd = getpass.getpass("\nVisTrails DB password for user %s:"%config['user'])
                self._user = config['user']
                self._passwd = passwd
                dbconfig = {'host': self._host,
                          'port': int(self._port),
                          'user': self._user,
                          'passwd': self._passwd,
                          'db': self._db
                          }
                try:
                    test_db_connection(dbconfig)
                    config['succeeded'] = True
                    config['passwd'] = self._passwd
                except VistrailsDBException, e:
                    debug.critical('VisTrails DB Exception', e)
                    config['succeeded'] = False
            
            if config['succeeded'] == True:
                self._host = config['host']
                self._port = config['port']
                self._db = config['db']
                self._user = config['user']
                self._passwd = config['passwd']
                self.ext_connection_id = self.set_connection_info(**config)
                return True
            return False
        return False
        
    def find_connection_info(self, host, port, db):
        """find_connection_info(host:str, port: int, db: str) -> dict
        Returns complete info of a connection with the given parameters

        """
        id = self.__list.find_db_connection(host,port,db)
        if id != -1:
            return self.get_connection_info(id)
        else:
            return None
    
    def get_connection_info(self, id):
        """get_connection_info(id: int) -> dict
        Returns info of ExtConnection """
        conn = self.__list.get_connection(id)
        if conn is not None:
            succeeded = False
            key = str(conn.id) + "." + conn.name + "." + conn.host
            passwd = DBLocator.keyChain.get_key(key)
            config = {'host': conn.host,
                      'port': conn.port,
                      'user': conn.user,
                      'passwd': passwd}
            try:
                test_db_connection(config)
                succeeded = True
            except VistrailsDBException:
                succeeded = False
                
            config['id'] = conn.id
            config['name'] = conn.name
            config['db'] = conn.database
            config['succeeded'] = succeeded
        else:
            config = None
        return config
    
    def set_connection_info(self, *args, **kwargs):
        """set_connection_info(id: int, name: str, host: str, port:int,
                     user:str, passwd:str, db:str) -> None
        If the connection exists it will update it, else it will add it

        """
        id = kwargs["id"]
        name = kwargs["name"]
        host = kwargs["host"]
        port = kwargs["port"]
        user = kwargs["user"]
        passwd = kwargs["passwd"]
        db = kwargs["db"]

        conn = DBConnection(id=id,
                            name=name,
                            host=host,
                            port=port,
                            user=user,
                            passwd='',
                            database=db,
                            dbtype='MySQL')
        
        if self.__list.has_connection(id):   
            self.__list.set_connection(id,conn)
        else:
            if conn.id == -1:
                conn.id = self.__list.get_fresh_id()
            self.__list.add_connection(conn)
        key = str(conn.id) + "." + conn.name + "." + conn.host
        DBLocator.keyChain.set_key(key,passwd)
        return conn.id
    
    ##########################################################################

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return (self._host == other._host and
                self._port == other._port and
                self._db == other._db and
                self._user == other._user and
                #self._name == other._name and
                long(self._obj_id) == long(other._obj_id) and
                self._obj_type == other._obj_type)

    ##########################################################################

    @staticmethod
    def prompt_autosave(parent_widget):
        return True

    @staticmethod
    def load_from_gui(parent_widget, obj_type):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_load_db_locator_from_gui(parent_widget, obj_type)

    @staticmethod
    def save_from_gui(parent_widget, obj_type, locator=None):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_save_db_locator_from_gui(parent_widget, obj_type,
                                                   locator)

    @staticmethod
    def from_xml(node, include_name=False):
        locator = _DBLocator.from_xml(node, include_name)
        locator.__class__ = DBLocator
        return locator
    
    @staticmethod
    def convert(locator):
        locator.__class__ = DBLocator
        locator.__list = ExtConnectionList.getInstance(
                                                   default_connections_file())

class ZIPFileLocator(_ZIPFileLocator, CoreLocator):

    def __init__(self, filename, **kwargs):
        _ZIPFileLocator.__init__(self, filename, **kwargs)

    def load(self, klass=None):
        from vistrails.core.vistrail.vistrail import Vistrail
        if klass is None:
            klass = Vistrail
        save_bundle = _ZIPFileLocator.load(self, klass.vtType)
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        return save_bundle

    def save(self, save_bundle):
        save_bundle = _ZIPFileLocator.save(self, save_bundle, False)
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        return save_bundle

    def save_as(self, save_bundle, version=None):
        save_bundle = _ZIPFileLocator.save(self, save_bundle, True, version)
        for obj in save_bundle.get_db_objs():
            klass = self.get_convert_klass(obj.vtType)
            klass.convert(obj)
            obj.locator = self
        # Need to update thumbnail cache since files have moved
        ThumbnailCache.getInstance().add_entries_from_files(save_bundle.thumbnails)
        return save_bundle

    ##########################################################################

    def __eq__(self, other):
        if not isinstance(other, ZIPFileLocator):
            return False
        return self._name == other._name

    ##########################################################################

    @staticmethod
    def prompt_autosave(parent_widget):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_autosave_prompt(parent_widget)

    @staticmethod
    def load_from_gui(parent_widget, obj_type):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_load_file_locator_from_gui(parent_widget, obj_type)

    @staticmethod
    def save_from_gui(parent_widget, obj_type, locator=None):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_save_file_locator_from_gui(parent_widget, obj_type,
                                                         locator)

class FileLocator(CoreLocator):
    def __new__(self, filename=None, **kwargs):
        if filename:
            if filename.endswith('.vt'):
                return ZIPFileLocator(filename, **kwargs)
            elif filename.endswith('.vtl'):
                return FileLocator.from_link_file(filename)
            else:
                return XMLFileLocator(filename, **kwargs)
        else:
            #return class based on default file type
            if vistrails_default_file_type() == '.vt':
                return ZIPFileLocator
            else:
                return XMLFileLocator

    @staticmethod
    def prompt_autosave(parent_widget):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_autosave_prompt(parent_widget)
    
    @staticmethod
    def load_from_gui(parent_widget, obj_type):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_load_file_locator_from_gui(parent_widget, obj_type)

    @staticmethod
    def save_from_gui(parent_widget, obj_type, locator=None):
        import vistrails.gui.extras.core.db.locator as db_gui
        return db_gui.get_save_file_locator_from_gui(parent_widget, obj_type,
                                                         locator)

    @staticmethod
    def parse(element):
        """ parse(element) -> XMLFileLocator
        Parse an XML object representing a locator and returns an
        XMLFileLocator or a ZIPFileLocator object.

        """
        if str(element.getAttribute('type')) == 'file':
            for n in element.childNodes:
                if n.localName == "name":
                    filename = str(n.firstChild.nodeValue).strip(" \n\t")
                    return FileLocator(filename)
            return None
        else:
            return None
    
    #ElementTree port
    @staticmethod
    def from_xml(node):
        """from_xml(node:ElementTree.Element) -> XMLFileLocator or None
        Parse an XML object representing a locator and returns a
        XMLFileLocator or a ZIPFileLocator object."""
        if node.tag != 'locator':
            return None
        type_ = node.get('type', '')
        if str(type_) == 'file':
            for child in node.getchildren():
                if child.tag == 'name':
                    filename = child.text.encode('latin-1').strip()
                    return FileLocator(filename)
        return None
    
    @staticmethod
    def from_link_file(filename):
        """from_link_file(filename: str) -> DBLocator
        This will parse a '.vtl' file and  will create a DBLocator. .vtl files
        are vistrail link files and they are used to point vistrails to open
        vistrails from the database on the web. """
        def convert_from_str(value,type):
            def bool_conv(x):
                s = str(x).upper()
                if s == 'TRUE':
                    return True
                if s == 'FALSE':
                    return False
            
            if value is not None:
                if type == 'str':
                    return str(value)
                elif value.strip() != '':
                    if type == 'long':
                        return long(value)
                    elif type == 'float':
                        return float(value)
                    elif type == 'int':
                        return int(value)
                    elif type == 'bool':
                        return bool_conv(value)
                    elif type == 'base64':
                        return base64.b64decode(value)
            return None
        
        def guess_extension_from_contents(contents):
            if contents.startswith("<vistrail"):
                return ".xml"
            else:
                return ".vt"
            
        tree = ElementTree.parse(filename)
        node = tree.getroot()
        if node.tag != 'vtlink':
            return None
        #read attributes
        data = node.get('host', None)
        host = convert_from_str(data, 'str')
        data = node.get('port', None)
        port = convert_from_str(data,'int')
        data = node.get('database', None)
        database = convert_from_str(data,'str')
        data = node.get('vtid')
        vt_id = convert_from_str(data, 'int')
        data = node.get('version')
        version = convert_from_str(data, 'str')
        data = node.get('tag')
        tag = convert_from_str(data, 'str')
        data = node.get('execute')
        execute = convert_from_str(data, 'bool')
        data = node.get('showSpreadsheetOnly')
        showSpreadsheetOnly = convert_from_str(data, 'bool')
        data = node.get('url', None)
        url = convert_from_str(data,'str')
        data = node.get('vtcontent', None)
        vtcontent = convert_from_str(data,'base64')
        data = node.get('filename', None)
        vtname = convert_from_str(data, 'str')
        data = node.get('forceDB',None)
        forceDB = convert_from_str(data,'bool')
        data = node.get('mashuptrail', None)
        mashuptrail = convert_from_str(data, 'str')
        data = node.get('mashupVersion', None)
        mashupVersion = convert_from_str(data, 'int')
        data = node.get('parameterExploration', None)
        parameterExploration = convert_from_str(data, 'int')
        
        #if execute is False, we will show the builder too
        if showSpreadsheetOnly and not execute:
            showSpreadsheetOnly = False
        try:
            version = int(version)
        except (ValueError, TypeError):
            pass

        if tag is None:
            tag = ''
            
        ## execute and showSpreadsheetOnly should be written to the current
        ## configuration
        config = get_vistrails_configuration()
        config.execute = execute
        config.showWindow = not showSpreadsheetOnly
        if not forceDB:
            if vtcontent is not None:
                if url is not None:
                    basename = url.split('/')[-1]
                    base,ext = os.path.splitext(basename)
                    dirname = os.path.dirname(filename)
                    fname = os.path.join(dirname,basename)
                else:
                    basename = os.path.basename(filename)
                    base,ext = os.path.splitext(basename)
                    ext = guess_extension_from_contents(vtcontent)
                    dirname = os.path.dirname(filename)
                    fname = os.path.join(dirname,"%s%s"%(base,ext))
                create_file = True
                if os.path.exists(fname): #file was extracted before
                    create_file = False
                    oldf = open(fname)
                    oldcontents = oldf.read()
                    if oldcontents != vtcontent:
                        import vistrails.gui.extras.core.db.locator as db_gui
                        (overwrite, newname) = \
                                 db_gui.ask_to_overwrite_file(None, 'vistrail')
                        create_file = True
                        if newname:
                            fname = newname
                        elif overwrite == False:
                            i=1
                            while os.path.exists(fname):
                                newbase = "%s_%s%s" % (base, i, ext)
                                fname = os.path.join(dirname,newbase)
                                i+=1
                        
                if create_file:
                    f = open(fname,'wb')
                    f.write(vtcontent)
                    f.close()
                return FileLocator(fname, version_node=version, version_tag=tag,
                                   mashuptrail=mashuptrail,
                                   mashupVersion=mashupVersion,
                                   parameterExploration=parameterExploration)
        if host is not None:
            user = ""
            passwd = ""
            
            return DBLocator(host, port, database,
                             user, passwd, None, obj_id=vt_id,
                             obj_type='vistrail',connection_id=None,
                             version_node=version, version_tag=tag,
                             mashuptrail=mashuptrail,
                             mashupVersion=mashupVersion,
                             parameterExploration=parameterExploration)
        elif vtname is not None:
            if os.path.dirname(vtname) == '':
                #check if file exists in the same directory as the .vtl file
                dirname = os.path.dirname(filename)
                newvtname = os.path.join(dirname,vtname)
                if os.path.exists(newvtname):
                    vtname = newvtname
            #check for magic strings
            if "@examples" in vtname:
                vtname=vtname.replace("@examples", vistrails_examples_directory())
            return FileLocator(vtname, version_node=version, version_tag=tag,
                               mashuptrail=mashuptrail,
                               mashupVersion=mashupVersion,
                               parameterExploration=parameterExploration)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.query import Query
from combined import CombinedSearch

class MultipleSearch(Query):
    # vistrails_to_check { Vistrail: set(version_ids) }
    def __init__(self, search_str=None, pipeline=None, entities_to_check={},
                 use_regex=False):
        self.entities_to_check = entities_to_check
        self.queryPipeline = pipeline
        self.search_str = search_str
        self.use_regex = use_regex
        self.queries = {}
        self.queries_by_vistrail = {}
        self.cur_vistrail = None

    def setCurrentVistrail(self, vistrail):
        self.cur_vistrail = vistrail

    def run(self):
        for entity, versions_to_check in self.entities_to_check.iteritems():
            query = CombinedSearch(self.search_str, self.queryPipeline, 
                                   versions_to_check, self.use_regex)
            query.run(entity.vistrail, '')
            self.queries[entity] = query
            self.queries_by_vistrail[entity.vistrail] = query

    def match(self, vistrail, action):
        self.setCurrentVistrail(vistrail)
        query = self.queries_by_vistrail[vistrail]
        return query.match(vistrail, action)

    def matchModule(self, version_id, module):
        query = self.queries_by_vistrail[self.cur_vistrail]
        return query.matchModule(version_id, module)
    
    def getResultEntities(self):
        result_entities = []
        for entity, query in self.queries.iteritems():
            versions_to_check = self.entities_to_check[entity]
            result_entity = query.getResultEntity(entity.vistrail, 
                                                  versions_to_check)
            if result_entity is not None:
                # needed for workspace results that are temporary...
                result_entity._window = entity._window
                result_entities.append(result_entity)
        return result_entities



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import base64
import os
import vistrails.core.db.action
from vistrails.core.modules.basic_modules import File, Boolean, String, Directory
from vistrails.core.modules.vistrails_module import Module, ModuleError, NotCacheable
from vistrails.core.vistrail.vistrail import Vistrail
from vistrails.db.services.locator import DBLocator
from vistrails.core.system import get_elementtree_library
from vistrails.db.services import io
from vistrails.db.versions import currentVersion

ElementTree = get_elementtree_library()

class VtlFileCreator(NotCacheable, Module):
    """This Module creates a vtl file for the workflow where it is
    present.
    By default it generates a string with an <vtlink> </vtlink> XML
    element or it can write the vtl file to disk if filename is
    provided. If directory is provided, the final filename will be
    the concatenation of directory and filename.
    
    Other input ports:
    
    execute: Boolean
        tells VisTrails to execute the workflow when the vtl is open
        
    showSpreadsheetOnly: Boolean
        tells VisTrails to hide the builder window and show only the
        spreadsheet when the vtl is open
    
    embedWorkflow: Boolean
        it will embed a vistrail encoded as base64 containing only 
        the workflow where the module is. It can be loaded by 
        VisTrails.
        
    """
    def __init__(self):
        Module.__init__(self)
        self.locator = None
        self.version = -1
        self.pipeline = None
        self.execute = False
        self.embedWorkflow = False
        self.showSpreadsheetOnly = False
        self.forceDB = False
        
    def get_locator_and_version(self):
        self.locator = self.moduleInfo['locator']
        self.version = self.moduleInfo['version']
        self.pipeline = self.moduleInfo['pipeline']
        
        if self.locator is None:
            raise ModuleError(self, 'could not get the locator for this pipeline')
        if self.version is None:
            raise ModuleError(self, 'could not get the version number of this peline')
    
    @staticmethod
    def generate_vtl(locator,version,pipeline,execute=False,forceDB=False,
                     showSpreadsheetOnly=False,embedWorkflow=False):
        """generate_vtl(locator:DBLocator or XMLLocator,
                        version: str, pipeline:Pipeline, execute:boolean,
                        forceDB:boolean, showspreadsheetOnly:boolean,
                        embedWorkflow: boolean) -> str
           It generates the contents of a .vtl file with the information
           given.
        """
        node = ElementTree.Element('vtlink')
        
        if isinstance(locator, DBLocator):
            node.set('host', str(locator.host))
            node.set('port', str(locator.port))
            node.set('database', str(locator.db))
            node.set('vtid', str(locator.obj_id))
        elif locator is not None:
            node.set('filename', str(locator.name))
            
        node.set('version', str(version))    
        node.set('execute', str(execute))
        node.set('forceDB', str(forceDB))
        node.set('showSpreadsheetOnly', str(showSpreadsheetOnly))
            
        if embedWorkflow == True:
            vistrail = Vistrail()
            action_list = []
            for module in pipeline.module_list:
                action_list.append(('add', module))
            for connection in pipeline.connection_list:
                action_list.append(('add', connection))
            action = vistrails.core.db.action.create_action(action_list)
            vistrail.add_action(action, 0L)
            vistrail.addTag("Imported workflow", action.id)
            if not forceDB:
                node.set('version', str(action.id))
            if not vistrail.db_version:
                vistrail.db_version = currentVersion
            pipxmlstr = io.serialize(vistrail)
            vtcontent = base64.b64encode(pipxmlstr)
            node.set('vtcontent',vtcontent)
            
        return ElementTree.tostring(node)
            
            
    def compute(self):
        self.get_locator_and_version()
        
        if self.has_input('execute'):
            self.execute = self.get_input('execute')
        
        if self.has_input('forceDB'):
            self.forceDB = self.get_input('forceDB')
        
        if self.has_input('showSpreadsheetOnly'):
            self.showSpreadsheetOnly = self.get_input('showSpreadsheetOnly')
            
        if self.has_input('embedWorkflow'):
            self.embedWorkflow = self.get_input('embedWorkflow')
        
        xmlstring = self.generate_vtl(self.locator,self.version,self.pipeline,
                                      self.execute,self.forceDB,
                                      self.showSpreadsheetOnly,self.embedWorkflow)
        
        if self.has_input('filename'):
            filename = self.get_input('filename')
            if self.has_input('directory'):
                directory = self.get_input('directory').name
                filename = os.path.join(directory,filename)
            file_ = open(filename,'w')
            file_.write(xmlstring)
            file_.close()
        
        self.set_output("xmlstring", xmlstring)
        
    _input_ports = [('execute', Boolean, True), 
                    ('showspreadsheetOnly', Boolean, True),
                    ('embedWorkflow', Boolean, True),
                    ('forceDB', Boolean, True), 
                    ('filename', String), 
                    ('directory', Directory)]
    
    _output_ports = [('xmlstring', String)]

_modules = [VtlFileCreator]



#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
#FIXME !!! DOES NOT MERGE EXECUTION LOGS STORED IN VT FILE !!!

import sys
if '../vistrails' not in sys.path:
    sys.path.append('../vistrails')

import vistrails.db.services.vistrail
from vistrails.db.domain import DBModule, DBConnection, DBPort, DBFunction, \
    DBParameter, DBLocation, DBPortSpec, DBTag, DBAnnotation, DBVistrail, \
    DBAction
from vistrails.db.services import io

def main(out_fname, in_fnames):
    """main(out_fname: str, in_fnames: list<str>) -> None
    Combines all vistrails specified in in_fnames into a single vistrail
    by renumbering their ids

    """
    # FIXME this breaks when you use abstractions!
    (save_bundle, vt_save_dir) = io.open_bundle_from_zip_xml(DBVistrail.vtType, in_fnames[0])
    for in_fname in in_fnames[1:]:
        (new_save_bundle, new_save_dir) = io.open_bundle_from_zip_xml(DBVistrail.vtType, in_fname)
        vistrails.db.services.vistrail.merge(save_bundle, new_save_bundle, "", True, vt_save_dir, new_save_dir)
    io.save_bundle_to_zip_xml(save_bundle, out_fname)
    
if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: %s [output filename] [list of input vistrails]" % \
            sys.argv[0]
        sys.exit(0)
    main(sys.argv[1], sys.argv[2:])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from itertools import izip, chain
import re

import vtk

from .class_tree import ClassTree
from .specs import ClassSpec, SpecList, \
                   ClassInputPortSpec as InputPortSpec, \
                   ClassOutputPortSpec as OutputPortSpec
from .vtk_parser import VTKMethodParser

parser = VTKMethodParser()

disallowed_classes = set(
    [
        'simplewrapper', # ticket 464: VTK 5.10 on OpenSuSE needs this
        'vtkEventQtSlotConnect', # VTK 5.10.1 OpenSuSE (uses QObject)
        'vtkQtView', # VTK 5.10.1 OpenSuSE (uses QWidget)
        'vtkCriticalSection',
        'vtkDataArraySelection',
        'vtkDebugLeaks',
        'vtkDirectory',
        'vtkDynamicLoader',
        'vtkFunctionParser',
        'vtkGarbageCollector',
        'vtkHeap',
        'vtkInformationKey',
        'vtkInstantiator',
        'vtkLogLookupTable', # use vtkLookupTable.SetScaleToLog10() instead
        'vtkMath',
        'vtkModelMetadata',
        'vtkMultiProcessController',
        'vtkMutexLock',
        'vtkOutputWindow',
        'vtkPriorityQueue',
        'vtkQtInitialization',
        'vtkReferenceCount',
        'vtkRenderWindowCollection',
        'vtkRenderWindowInteractor',
        'vtkTesting',
        'vtkWindow',
        'vtkContext2D',       #Not working for VTK 5.7.0
        'vtkPLYWriter',       #Not working for VTK 5.7.0.
        'vtkBooleanTexture',  #Not working for VTK 5.7.0
        'vtkImageMaskBits',   #Not working for VTK 5.7.0
        'vtkHardwareSelector',#Not working for VTK 5.7.0
        'vtkOpenGLExtensionManager',

        # these show up with new parse
        'vtkAbstractContextBufferId',
        'vtkAbstractElectronicData',
        'vtkCallbackCommand',
        'vtkImageComplex',
        'vtkInformationDataObjectKey',
        'vtkInformationDoubleKey',
        'vtkInformationDoubleVectorKey',
        'vtkInformationIdTypeKey',
        'vtkInformationInformationKey',
        'vtkInformationInformationVectorKey',
        'vtkInformationIntegerKey',
        'vtkInformationIntegerPointerKey',
        'vtkInformationIntegerVectorKey',
        'vtkInformationKeyVectorKey',
        'vtkInformationObjectBaseKey',
        'vtkInformationRequestKey',
        'vtkInformationStringKey',
        'vtkInformationStringVectorKey',
        'vtkInformationUnsignedLongKey',
        'vtkRenderWindow',
        'vtkShaderProgram2',
        'vtkShadowMapBakerPassLightCameras',
        'vtkShadowMapBakerPassTextures',
        'vtkTDxMotionEventInfo',
        'vtkVolumeRayCastDynamicInfo',
        'vtkVolumeRayCastStaticInfo',
        # For VTK 5.8
        'vtkMPICommunicatorOpaqueComm',
        # For VTK 6
        'vtkBlueObeliskData',
        'vtkSocketController',
        'vtkMPIController',
        'vtkInformationVariantVectorKey',
        'vtkInformationVariantKey',
        'QImage',
        'vtkPLOT3DReader',
        # For VTK 6.2
        'QuantileDefinitionType'
    ])

disallowed_modules = set(
    [
        'vtkGeoAlignedImageCache',
        'vtkGeoTerrainCache',
        'vtkMPIGroup'
    ])

def create_module(base_cls_name, node):
    """create_module(base_cls_name: String, node: TreeNode) -> [ModuleSpec]

    Construct a module spec that inherits from base_cls_name with
    specification from node.

    """
    if node.name in disallowed_modules: return []
    if node.name == 'int': return [] #enum
    def obsolete_class_list():
        lst = []
        items = ['vtkInteractorStyleTrackball',
                 'vtkStructuredPointsGeometryFilter',
                 'vtkConstrainedPointHandleRepresentation',
                 'vtkRenderViewBase',
                 'vtkRenderView']
        def try_to_add_item(item):
            try:
                lst.append(getattr(vtk, item))
            except AttributeError:
                pass
        for item in items:
            try_to_add_item(item)
        return lst

    obsolete_list = obsolete_class_list()

    def is_abstract():
        """is_abstract tries to instantiate the class. If it's
        abstract, this will raise."""
        # Consider obsolete classes abstract
        if node.klass in obsolete_list:
            return True
        try:
            getattr(vtk, node.name)()
        except (TypeError, NotImplementedError): # VTK raises type error on abstract classes
            return True
        return False

    try:
        node.klass.__doc__.decode('latin-1')
    except UnicodeDecodeError:
        print "ERROR decoding docstring", node.name
        raise

    input_ports, output_ports = get_ports(node.klass)
    output_ports = list(output_ports) # drop generator

    cacheable = (issubclass(node.klass, vtk.vtkAlgorithm) and
                 (not issubclass(node.klass, vtk.vtkAbstractMapper))) or \
                issubclass(node.klass, vtk.vtkScalarTree)

    is_algorithm = issubclass(node.klass, vtk.vtkAlgorithm)
    tempfile = '_set_tempfile' if issubclass(node.klass, vtk.vtkWriter) else None
    callback = '_set_callback' if is_algorithm else None
    methods_last = hasattr(node.klass, 'SetRenderWindow')

    module_spec = ClassSpec(node.name, base_cls_name, node.name,
                            node.klass.__doc__.decode('latin-1'), callback,
                            tempfile, cacheable, input_ports, output_ports,
                            compute='Update', cleanup='_cleanup',
                            methods_last=methods_last, abstract=is_abstract())

    module_specs = [module_spec]
    for child in node.children:
        if child.name in disallowed_classes:
            continue
        module_specs.extend(create_module(node.name, child))

    return module_specs

def get_doc(cls, port_name):
    f = re.match(r"(.*)_\d+$", port_name)
    if f:
        name = f.group(1)
    else:
        name = port_name

    doc = getattr(cls, name).__doc__
    # Remove all the C++ function signatures.
    idx = doc.find('\n\n')
    if idx > 0:
        doc = doc[idx+2:]
    return doc

def prune_signatures(cls, name, signatures, output=False):
    """prune_signatures tries to remove redundant signatures to reduce
    overloading. It _mutates_ the given parameter.

    It does this by performing several operations:

    1) It compares a 'flattened' version of the types
    against the other 'flattened' signatures. If any of them match, we
    keep only the 'flatter' ones.

    A 'flattened' signature is one where parameters are not inside a
    tuple.

    2) We explicitly forbid a few signatures based on modules and names

    """
    # yeah, this is Omega(n^2) on the number of overloads. Who cares?

    def flatten(type_):
        if type_ is None:
            return []
        def convert(entry):
            if isinstance(entry, tuple):
                return list(entry)
            elif isinstance(entry, str):
                return [entry]
            else:
                result = []
                first = True
                lastList = True
                for e in entry:
                    if (isinstance(e, list)):
                        if lastList == False: result[len(result)] = result[len(result)] + ']'
                        aux = e
                        aux.reverse()
                        aux[0] = '[' + aux[0]
                        aux[-1] = aux[-1] + ']'
                        result.extend(aux)
                        lastList = True
                    else:
                        if first: e = '[' + e
                        result.append(e)
                        lastList = False
                        first = False
                return result
        result = []
        for entry in type_:
            result.extend(convert(entry))
        return result
    flattened_entries = [flatten(sig[1]) for
                         sig in signatures]
    def hit_count(entry):
        result = 0
        for entry in flattened_entries:
            if entry in flattened_entries:
                result += 1
        return result
    hits = [hit_count(entry) for entry in flattened_entries]

    def forbidden(flattened, hit_count, original):
        if (issubclass(cls, vtk.vtk3DWidget) and
            name == 'PlaceWidget' and
            flattened == []):
            return True
        # We forbid this because addPorts hardcodes this but
        # SetInputArrayToProcess is an exception for the InfoVis
        # package
        if (cls == vtk.vtkAlgorithm and
            name!='SetInputArrayToProcess'):
            return True
        return False

    # This is messy: a signature is only allowed if there's no
    # explicit disallowing of it. Then, if it's not overloaded,
    # it is also allowed. If it is overloaded and not the flattened
    # version, it is pruned. If these are output ports, there can be
    # no parameters.

    def passes(flattened, hit_count, original):
        if forbidden(flattened, hit_count, original):
            return False
        if hit_count == 1:
            return True
        if original[1] is None:
            return True
        if output and len(original[1]) > 0:
            return False
        if hit_count > 1 and len(original[1]) == len(flattened):
            return True
        return False

    signatures[:] = [original for (flattened, hit_count, original)
                     in izip(flattened_entries,
                             hits,
                             signatures)
                     if passes(flattened, hit_count, original)]

    #then we remove the duplicates, if necessary
    unique_signatures = []

    #Remove the arrays and tuples inside the signature
    #  in order to transform it in a single array
    #Also remove the '[]' from the Strings
    def removeBracts(signatures):
        result = []
        stack = list(signatures)
        while (len(stack) != 0):
            curr = stack.pop(0)
            if (isinstance(curr, str)):
                c = curr.replace('[', '')
                c = c.replace(']', '')
                result.append(c)
            elif (curr == None):
                result.append(curr)
            elif (isinstance(curr, list)):
                curr.reverse()
                for c in curr: stack.insert(0, c)
            elif (isinstance(curr, tuple)):
                cc = list(curr)
                cc.reverse()
                for c in cc: stack.insert(0, c)
            else:
                result.append(curr)
        return result

    unique2 = []
    for s in signatures:
        aux = removeBracts(s)
        if not unique2.count(aux):
            unique_signatures.append(s)
            unique2.append(aux)
    signatures[:] = unique_signatures

file_name_pattern = re.compile('.*FileName$')
set_file_name_pattern = re.compile('Set.*FileName$')

def resolve_overloaded_name(name, ix, signatures):
    # VTK supports static overloading, VisTrails does not. The
    # solution is to check whether the current function has
    # overloads and change the names appropriately.
    if len(signatures) == 1:
        return name
    else:
        return name + '_' + str(ix+1)

type_map_dict = {'int': "basic:Integer",
                 'long': "basic:Integer",
                 'float': "basic:Float",
                 'char*': "basic:String",
                 'char *': "basic:String",
                 'string': "basic:String",
                 'char': "basic:String",
                 'const char*': "basic:String",
                 'const char *': "basic:String",
                 '[float': "basic:Float",
                 'float]': "basic:Float",
                 '[int': "basic:Integer",
                 'int]': "basic:Integer",
                 'bool': "basic:Boolean",
                 'unicode': 'basic:String'}

type_map_values = set(type_map_dict.itervalues())
# ["basic:Integer", "basic:Float", "basic:String", "basic:Boolean"]

def get_port_types(name):
    """ get_port_types(name: str) -> str
    Convert from C/C++ types into VisTrails port type

    """
    if isinstance(name, tuple) or isinstance(name, list):
        return [get_port_types(x) for x in name]
    if name in type_map_dict:
        return type_map_dict[name]
    else:
        if name is not None and name.strip():
            #if not name.startswith("vtk"):
            #    print "RETURNING RAW TYPE:", name
            return name
    return None

def is_type_allowed(t):
    if isinstance(t, list):
        return all(is_type_allowed(sub_t) for sub_t in t)
    if t is None:
        return False
    if t == "tuple" or t == "function":
        return False
    return t not in disallowed_classes

def get_algorithm_ports(cls):
    """ get_algorithm_ports(cls: class) -> None
    If module is a subclass of vtkAlgorithm, this function will add all
    SetInputConnection([id],[port]) and GetOutputPort([id]) as
    SetInputConnection{id}([port]) and GetOutputPort{id}.

    """
    input_ports = []
    output_ports = []

    if issubclass(cls, vtk.vtkAlgorithm):
        # We try to instantiate the class here to get the number of
        # ports and to avoid abstract classes
        try:
            instance = cls()
        except TypeError:
            pass
        else:
            for i in xrange(instance.GetNumberOfInputPorts()):
                port_name = "SetInputConnection%d" % i
                port_spec = InputPortSpec(name=port_name,
                                          method_name="SetInputConnection",
                                          port_type="vtkAlgorithmOutput",
                                          docstring=get_doc(cls,
                                                        "SetInputConnection"),
                                          show_port=True,
                                          prepend_params=[i])
                input_ports.append(port_spec)
            for i in xrange(instance.GetNumberOfOutputPorts()):
                port_name = "GetOutputPort%d" % i
                port_spec = OutputPortSpec(name=port_name,
                                           method_name="GetOutputPort",
                                           port_type="vtkAlgorithmOutput",
                                           docstring=get_doc(cls,
                                                             "GetOutputPort"),
                                           show_port=True)
                output_ports.append(port_spec)

    return input_ports, output_ports

disallowed_get_ports = set([
    'GetClassName',
    'GetErrorCode',
    'GetNumberOfInputPorts',
    'GetNumberOfOutputPorts',
    'GetOutputPortInformation',
    'GetTotalNumberOfInputConnections',
    ])

def get_get_ports(cls, get_list):
    output_ports = []
    for name in get_list:
        if name in disallowed_get_ports:
            continue
        method = getattr(cls, name)
        signatures = parser.get_method_signature(method)
        if len(signatures) > 1:
            prune_signatures(cls, name, signatures, output=True)
        for ix, getter in enumerate(signatures):
            if getter[1]:
                #print ("Can't handle getter %s (%s) of class %s: Needs input "
                #       "to get output" % (ix+1, name, cls.__name__))
                continue
            if len(getter[0]) != 1:
                #print ("Can't handle getter %s (%s) of class %s: More than a "
                #       "single output" % (ix+1, name, cls.__name__))
                continue
            port_type = get_port_types(getter[0][0])
            if is_type_allowed(port_type):
                n = resolve_overloaded_name(name[3:], ix, signatures)
                port_spec = OutputPortSpec(name=n,
                                           method_name=name,
                                           port_type=port_type,
                                           show_port=False,
                                           docstring=get_doc(cls, name))
                output_ports.append(port_spec)
    return [], output_ports

disallowed_get_set_ports = set(['ReferenceCount',
                                'InputConnection',
                                'OutputPort',
                                'Progress',
                                'ProgressText',
                                'InputArrayToProcess',
                                ])

color_ports = set(["DiffuseColor", "Color", "AmbientColor", "SpecularColor",
                   "EdgeColor", "Background", "Background2"])

to_vtk6_names = {'AddInput':  'AddInputData',
                 'SetInput':  'SetInputData',
                'AddSource': 'AddSourceData',
                'SetSource': 'SetSourceData'}
def get_vtk6_name(cls, name):
    # Return SetInputData for SetInput etc.
    if name == 'AddInput' and cls == vtk.vtkXYPlotActor:
        return 'AddDataSetInput'
    return to_vtk6_names.get(name, name)

# FIXME use defaults and ranges!
def get_get_set_ports(cls, get_set_dict):
    """get_get_set_ports(cls: class, get_set_dict: dict) -> None
    Convert all Setxxx methods of cls into input ports and all Getxxx
    methods of module into output ports

    Keyword arguments:
    cls          --- class
    get_set_dict --- the Set/Get method signatures returned by vtk_parser

    """

    input_ports = []
    output_ports = []
    for name in get_set_dict:
        if name in disallowed_get_set_ports:
            continue
        getter_name = 'Get%s' % name
        setter_name = 'Set%s' % name
        getter_method = getattr(cls, getter_name)
        setter_method = getattr(cls, setter_name)
        getter_sig = parser.get_method_signature(getter_method)
        setter_sig = parser.get_method_signature(setter_method)
        if len(getter_sig) > 1:
            prune_signatures(cls, getter_name, getter_sig, output=True)
        for order, getter in enumerate(getter_sig):
            if getter[1]:
                #print ("Can't handle getter %s (%s) of class %s: Needs input "
                #       "to get output" % (order+1, name, cls.__name__))
                continue
            if len(getter[0]) != 1:
                #print ("Can't handle getter %s (%s) of class %s: More than a "
                #       "single output" % (order+1, name, cls.__name__))
                continue
            port_type = get_port_types(getter[0][0])
            if is_type_allowed(port_type):
                if name in color_ports:
                    ps = OutputPortSpec(name=name,
                                        method_name=getter_name,
                                        port_type="basic:Color",
                                        show_port=False,
                                        docstring=get_doc(cls, getter_name))
                    input_ports.append(ps)
                else:
                    ps = OutputPortSpec(name=name,
                                        method_name=getter_name,
                                        port_type=port_type,
                                        show_port=False,
                                        docstring=get_doc(cls, getter_name))
                    output_ports.append(ps)

        if len(setter_sig) > 1:
            prune_signatures(cls, setter_name, setter_sig)
        docstring = get_doc(cls, setter_name)
        v = vtk.vtkVersion()
        version = [v.GetVTKMajorVersion(),
                   v.GetVTKMinorVersion(),
                   v.GetVTKBuildVersion()]
        if version < [6, 0, 0]:
            # Always use VTK6-style names for InputData-style types
            setter_name = get_vtk6_name(cls, setter_name)
            name = setter_name[3:]
        for ix, setter in enumerate(setter_sig):
            if setter[1] is None:
                continue

            # Wrap SetFileNames for VisTrails file access
            # FIXME add documentation
            if file_name_pattern.match(name):
                ps = InputPortSpec(name=name[:-4],
                                   method_name=setter_name,
                                   port_type="basic:File",
                                   show_port=True)
                input_ports.append(ps)
            # Wrap color methods for VisTrails GUI facilities
            # FIXME add documentation
            elif name in color_ports:
                ps = InputPortSpec(name=name,
                                   method_name=setter_name,
                                   port_type="basic:Color",
                                   show_port=False)
                input_ports.append(ps)
            # Wrap SetRenderWindow for exporters
            # FIXME Add documentation
            elif name == 'RenderWindow' and cls == vtk.vtkExporter:
                ps = InputPortSpec(name="vtkRenderer",
                                   port_type="vtkRenderer",
                                   show_port=True)
                input_ports.append(ps)
            else:
                n = resolve_overloaded_name(name, ix, setter_sig)
                port_types = get_port_types(setter[1])
                if is_type_allowed(port_types):
                    if len(setter[1]) == 1:
                        show_port = True
                        try:
                            show_port = port_types[0] not in type_map_values
                        except TypeError: # hash error
                            pass
                        port_types = port_types[0]
                    else:
                        show_port = False
                    ps = InputPortSpec(name=n,
                                       method_name=setter_name,
                                       port_type=port_types,
                                       show_port=show_port,
                                       docstring=docstring,
                                       depth=1)
                    input_ports.append(ps)

    return input_ports, output_ports

disallowed_toggle_ports = set(['GlobalWarningDisplay',
                               'Debug',
                               ])
def get_toggle_ports(cls, toggle_dict):
    """ get_toggle_ports(cls: class, toggle_dict: dict) -> None
    Convert all xxxOn/Off methods of module into boolean input ports

    Keyword arguments:
    module      --- Module
    toggle_dict --- the Toggle method signatures returned by vtk_parser

    """

    input_ports = []
    for name, default_val in toggle_dict.iteritems():
        if name in disallowed_toggle_ports:
            continue
        ps = InputPortSpec(name=name,
                           method_name=name, # With On/Off appended
                           method_type='OnOff',
                           port_type="basic:Boolean",
                           show_port=False,
                           defaults=[bool(default_val)],
                           docstring=get_doc(cls, name + "On"))
        input_ports.append(ps)
    return input_ports, []

disallowed_state_ports = set([('InputArray', 'Process')])
def get_state_ports(cls, state_dict):
    """ get_state_ports(cls: class, state_dict: dict) -> None
    Convert all SetxxxToyyy methods of module into input ports

    Keyword arguments:
    module     --- Module
    state_dict --- the State method signatures returned by vtk_parser

    """
    input_ports = []
    for name in state_dict:
        enum_values = []
        translations = {}
        method_name = "Set%sTo%s" % (name, state_dict[name][0][0])
        method_name_short = "Set%sTo" % name
        for mode in state_dict[name]:
            if (name, mode[0]) in disallowed_state_ports:
                continue
            if mode[0] in translations:
                if translations[mode[0]] != mode[1]:
                    raise Exception("Duplicate entry with different value")
                continue
            translations[mode[0]] = mode[1]
            enum_values.append(mode[0])

        ps = InputPortSpec(name=name,
                           method_name=method_name_short,
                           method_type='SetXToY',
                           port_type="basic:String",
                           entry_types=['enum'],
                           values=[enum_values],
                           show_port=False,
                           docstring=get_doc(cls, method_name))
        input_ports.append(ps)

    return input_ports, []


disallowed_other_ports = set(
    [
     'BreakOnError',
     'DeepCopy',
     'FastDelete',
     'HasObserver',
     'HasExecutive',
     'InvokeEvent',
     'IsA',
     'Modified',
     'NewInstance',
     'PrintRevisions',
     'RemoveAllInputs',
     'RemoveObserver',
     'RemoveObservers',
     'SafeDownCast',
#     'SetInputArrayToProcess',
     'ShallowCopy',
     'Update',
     'UpdateInformation',
     'UpdateProgress',
     'UpdateWholeExtent',
     # DAK: These are taken care of by s.upper() == s test
     # 'GUI_HIDE',
     # 'INPUT_ARRAYS_TO_PROCESS',
     # 'INPUT_CONNECTION',
     # 'INPUT_IS_OPTIONAL',
     # 'INPUT_IS_REPEATABLE',
     # 'INPUT_PORT',
     # 'INPUT_REQUIRED_DATA_TYPE',
     # 'INPUT_REQUIRED_FIELDS',
     # 'IS_INTERNAL_VOLUME',
     # 'IS_EXTERNAL_SURFACE',
     # 'MANAGES_METAINFORMATION',
     # 'POINT_DATA',
     # 'POINTS',
     # 'PRESERVES_ATTRIBUTES',
     # 'PRESERVES_BOUNDS',
     # 'PRESERVES_DATASET',
     # 'PRESERVES_GEOMETRY',
     # 'PRESERVES_RANGES',
     # 'PRESERVES_TOPOLOGY',
     # for VTK 6
     'SetMaterialProperties',
     ])


# FIXME deal with this in diff...
force_not_optional_port = set(
    ['ApplyViewTheme',
     ])

def get_other_ports(cls, other_list):
    """ addOtherPorts(cls: Module, other_list: list) -> None
    Convert all other ports such as Insert/Add.... into input/output

    Keyword arguments:
    cls        --- class
    other_dict --- any other method signatures that is not
                   Algorithm/SetGet/Toggle/State type

    """
    input_ports = []
    for name in other_list:
        # DAK: check for static methods as name.upper() == name
        if name in disallowed_other_ports or name.upper() == name:
            continue
        elif name=='CopyImportVoidPointer':
            # FIXME add documentation
            ps = InputPortSpec(name='CopyImportVoidString',
                               method_name='CopyImportVoidPointer',
                               port_type='basic:String',
                               show_port=True)

        # elif name[:3] in ['Add','Set'] or name[:6]=='Insert':
        else:
            method = getattr(cls, name)
            signatures = ""
            if not isinstance(method, int):
                signatures = parser.get_method_signature(method)
            if len(signatures) > 1:
                prune_signatures(cls, name, signatures)
            docstring = get_doc(cls, name)
            v = vtk.vtkVersion()
            version = [v.GetVTKMajorVersion(),
                       v.GetVTKMinorVersion(),
                       v.GetVTKBuildVersion()]
            if version < [6, 0, 0]:
                # Always use VTK6-style names for InputData-style types
                name = get_vtk6_name(cls, name)
            for (ix, sig) in enumerate(signatures):
                ([result], params) = sig
                port_types = get_port_types(params)
                if not (name[:3] in ['Add','Set'] or
                        name[:6]=='Insert' or
                        (port_types is not None and len(port_types) == 0) or
                        result is None):
                    continue
                if is_type_allowed(port_types):
                    n = resolve_overloaded_name(name, ix, signatures)
                    if n.startswith('Set'):
                        n = n[3:]
                    show_port = False
                    if len(port_types) < 1:
                        raise Exception("Shouldn't have empty input")
                    elif len(port_types) == 1:
                        if name[:3] in ['Add','Set'] or name[:6]=='Insert':
                            try:
                                show_port = port_types[0] not in type_map_values
                            except TypeError:
                                pass
                        port_types = port_types[0]
                    ps = InputPortSpec(name=n,
                                       method_name=name,
                                       port_type=port_types,
                                       show_port=show_port,
                                       docstring=docstring,
                                       depth=1)
                    input_ports.append(ps)
                elif result == None or port_types == []:
                    n = resolve_overloaded_name(name, ix, signatures)
                    ps = InputPortSpec(name=n,
                                       method_name=name,
                                       port_type='basic:Boolean',
                                       method_type='nullary',
                                       docstring=get_doc(cls, name),
                                       depth=1)
                    input_ports.append(ps)
    return input_ports, []

def get_custom_ports(cls):
    """ get_custom_ports(cls: Module) -> None
    Patch other ports needed to get a good wrapping

    Keyword arguments:
    cls        --- class

    """
    input_ports = []
    output_ports = []

    if cls == vtk.vtkAlgorithm:
        ps = InputPortSpec(name='AddInputConnection',
                           port_type='vtkAlgorithmOutput',
                           show_port=True,
                           docstring='Adds an input connection',
                           depth=1)
        input_ports.append(ps)
    # vtkWriters have a custom File port
    if cls in [vtk.vtkWriter, vtk.vtkImageWriter]:
        ps = OutputPortSpec(name='file',
                            port_type='basic:File',
                            show_port=True,
                            docstring='The written file')
        output_ports.append(ps)
    elif cls == vtk.vtkVolumeProperty:
        ps = InputPortSpec(name='TransferFunction',
                           method_name='SetTransferFunction',
                           port_type='TransferFunction',
                           docstring='Sets the transfer function to use')
        input_ports.append(ps)
    elif cls == vtk.vtkDataSet:
        ps = InputPortSpec(name='SetPointData',
                           method_name='PointData',
                           port_type='vtkPointData',
                           show_port=True,
                           docstring='Sets the point data')
        input_ports.append(ps)
        ps = InputPortSpec(name='SetCellData',
                           method_name='CellData',
                           port_type='vtkCellData',
                           show_port=True,
                           docstring='Sets the cell data')
        input_ports.append(ps)
    elif cls==vtk.vtkCell:
        ps = InputPortSpec(name='SetPointIds',
                           method_name='PointIds',
                           port_type='vtkIdList',
                           show_port=True,
                           docstring='Sets the point id list')
        input_ports.append(ps)
    elif cls==vtk.vtkMultiBlockPLOT3DReader:
        ps = OutputPortSpec(name='StructuredGrid',
                            method_name='FirstBlock',
                            port_type='vtkStructuredGrid',
                            show_port=True,
                            docstring='Returns .GetOutput().GetBlock(0)')
        output_ports.append(ps)

    return input_ports, output_ports


def get_ports(cls):
    """get_ports(cls: vtk class) -> None

    Search all metamethods of module and add appropriate ports

    """

    parser.parse(cls)
    ports_tuples = []
    ports_tuples.append(get_algorithm_ports(cls))
    ports_tuples.append(get_get_ports(cls, parser.get_get_methods()))
    ports_tuples.append(get_get_set_ports(cls, parser.get_get_set_methods()))
    ports_tuples.append(get_toggle_ports(cls, parser.get_toggle_methods()))
    ports_tuples.append(get_state_ports(cls, parser.get_state_methods()))
    ports_tuples.append(get_other_ports(cls, parser.get_other_methods()))
    ports_tuples.append(get_custom_ports(cls))

    zipped_ports = izip(*ports_tuples)
    input_ports = chain(*zipped_ports.next())
    output_ports = chain(*zipped_ports.next())
    return input_ports, output_ports

def parse(filename="vtk_raw.xml"):
    inheritance_graph = ClassTree(vtk)
    inheritance_graph.create()

    v = vtk.vtkVersion()
    version = [v.GetVTKMajorVersion(),
               v.GetVTKMinorVersion(),
               v.GetVTKBuildVersion()]
    if version < [5, 7, 0]:
        assert len(inheritance_graph.tree[0]) == 1
        base = inheritance_graph.tree[0][0]
        assert base.name == 'vtkObjectBase'

    specs_list = []
    if version < [5, 7, 0]:
        for child in base.children:
            if child.name in disallowed_classes:
                continue
            specs_list.extend(create_module("vtkObjectBase", child))
    else:
        for base in inheritance_graph.tree[0]:
            for child in base.children:
                if child.name in disallowed_classes:
                    continue
                specs_list.extend(create_module("vtkObjectBase", child))

    specs = SpecList(specs_list)
    specs.write_to_xml(filename)


if __name__ == '__main__':
    parse()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""This file contains classes working with cell helper widgets, i.e. toolbar,
resizer, etc.:
  CellHelpers
  CellResizer
  CellResizerConfig
"""

from __future__ import division

from PyQt4 import QtCore, QtGui


class CellResizerConfig(object):
    """
    CellResizerConfig can be used to config different parameters of
    the CellResizer widget such as shape, mask, pixmap, color, size,
    and cursor. By default, it has a black, triangular shape of size
    25x25. In order to change the shape, we have to override this
    class

    """
    def __init__(self, size=25, color=QtGui.QColor(0,0,0)):
        """ CellResizerConfig(size: int, color: QColor) -> CellResizerConfig
        Create mask and pixmap of a triangular shape with the specifc size
        and color

        """
        self.size = size
        self.transparentColor = QtGui.QColor(QtCore.Qt.blue)
        self.image = QtGui.QImage(size,size,QtGui.QImage.Format_RGB32)
        for i in xrange(size):
            for j in xrange(size):
                if i+j<size-1:
                    self.image.setPixel(i, j, self.transparentColor.rgb())
                else:
                    if i+j==size-1 or i==size-1 or j==size-1:
                        self.image.setPixel(i, j,
                                            QtGui.QColor(QtCore.Qt.white).rgb())
                    else:
                        self.image.setPixel(i, j, color.rgb())
        self.pixmapVar = self.maskVar = self.cursorVar = None

    def pixmap(self):
        """ pixmap() -> QPixmap
        Return the pixmap of the resizer shape

        """
        if not self.pixmapVar:
            self.pixmapVar = QtGui.QPixmap.fromImage(self.image)
        return self.pixmapVar

    def mask(self):
        """ mask() -> QRegion
        Return only the region of the resizer that will be shown on screen

        """
        if not self.maskVar:
            mask = self.pixmap().createMaskFromColor(self.transparentColor)
            self.maskVar = QtGui.QRegion(mask)
        return self.maskVar

    def cursor(self):
        """ cursor() -> QCursor
        Return the cursor that will be shown inside the resizer

        """
        return QtGui.QCursor(QtCore.Qt.SizeFDiagCursor)


class CellResizer(QtGui.QLabel):
    """
    CellResizer is a customized shape SizeGrip that stays on top of
    the widget, moving this size grip will end up resizing the
    corresponding cell in the table. This is different from QSizeGrip
    because it allows overlapping on top of the widget

    """
    def __init__(self, sheet, config=CellResizerConfig(), parent=None):
        """ CellResizer(sheet: SpreadsheetSheet,
                        config: subclass of CellResizerConfig,
                        parent: QWidget) -> CellResizer
        Initialize the size grip with the default triangular shape

        """
        QtGui.QLabel.__init__(self,sheet)
        self.setFixedSize(config.size, config.size)
        self.setPixmap(config.pixmap())
        self.setMask(config.mask())
        self.setCursor(config.cursor())
        self.setStatusTip("Left/Right-click drag to resize the underlying"
                          "cell or the whole spreadsheet ")
        self.setFocusPolicy(QtCore.Qt.NoFocus)
        self.sheet = sheet
        self.config = config
        self.resizeAll = False
        self.dragging = False
        self.lastPos = None
        self.row = -1
        self.col = -1
        self.hide()

    def setDragging(self,b):
        """ setDragging(b: boolean) -> None
        Set the resizer state to busy dragging

        """
        self.dragging = b

    def snapTo(self,row,col):
        """ snapTo(row, col) -> None
        Assign which row and column the resizer should be controlling

        """
        self.row = row
        self.col = col

    def adjustPosition(self, rect):
        """ adjustPosition(rect: QRect) -> None
        Adjust resizer position to be on the bottom-right corner of the cell

        """
        p = self.parent().mapFromGlobal(rect.topLeft())
        self.move(p.x()+rect.width()-self.width(),
                  p.y()+rect.height()-self.height())

    def mousePressEvent(self,e):
        """ mousePressEvent(e: QMouseEvent) -> None
        Handle Qt mouse press event to track if we need to resize
        either left or right mouse button is clicked

        """
        if self.col>=0:
            if e.button()==QtCore.Qt.LeftButton:
                self.resizeAll = False
                self.dragging = True
            if e.button()==QtCore.Qt.RightButton and not self.sheet.fitToWindow:
                self.resizeAll = True
                self.dragging = True
            self.lastPos = (e.globalX()-self.sheet.columnWidth(self.col),
                            e.globalY()-self.sheet.rowHeight(self.row))

    def mouseReleaseEvent(self,e):
        """ mouseReleaseEvent(e: QMouseEvent) -> None
        Handle Qt mouse release event to clean up all state

        """
        if (e.button()==QtCore.Qt.LeftButton or
            e.button()==QtCore.Qt.RightButton):
            self.dragging = False

    def mouseMoveEvent(self,e):
        """ mouseMoveEvent(e: QMouseEvent) -> None
        Interactively resize the corresponding column and row when the
        mouse moves

        """
        if self.dragging:
            hSize = self.sheet.columnWidth(self.col)
            vSize = self.sheet.rowHeight(self.row)
            hd = e.globalX() - self.lastPos[0] - hSize
            vd = e.globalY() - self.lastPos[1] - vSize

            # All sections should have the same size (Right-Click)
            if self.resizeAll:
                # Resize the columns first
                dS = int(hd // (self.col+1))
                mS = hd % (self.col+1)
                for i in xrange(self.sheet.columnCount()):
                    if i>self.col:
                        newValue = hSize+dS
                    else:
                        newValue = self.sheet.columnWidth(i)+dS+(i<mS)
                    self.sheet.setColumnWidth(i, newValue)
                # Then resize the rows
                dS = int(vd // (self.row+1))
                mS = vd % (self.row+1)
                for i in xrange(self.sheet.rowCount()):
                    if i>self.row:
                        newValue = vSize+dS
                    else:
                        newValue = self.sheet.rowHeight(i)+dS+(i<mS)
                    self.sheet.setRowHeight(i, newValue)

            # Only resize the corresponding column and row (Left-Click)
            else:
                self.sheet.setColumnWidth(self.col, hSize+hd)
                self.sheet.setRowHeight(self.row, vSize+vd)
            rect = self.sheet.getCellRect(self.row, self.col)
            rect.moveTo(self.sheet.viewport().mapToGlobal(rect.topLeft()))
            self.adjustPosition(rect)


class CellHelpers(object):
    """
    CellHelpers is a container include CellResizer that will shows up
    whenever the Ctrl key is hold down and the mouse hovers the cell.

    """
    def __init__(self, sheet, resizerInstance=None):
        """ CellHelpers(sheet: SpreadsheetSheet,
                        resizerInstance: CellResizer) -> CellHelpers
        Initialize with  a cell resizer

        """
        self.sheet = sheet
        self.resizer = resizerInstance
        self.row = -1
        self.col = -1

    def snapTo(self, row, col):
        """ snapTo(row: int, col: int) -> None
        Assign the resizer to the correct cell

        """
        if row>=0 and ((row!=self.row) or (col!=self.col)):
            self.hide()
            self.row = row
            self.col = col
            if self.resizer:
                self.resizer.snapTo(row,col)
            self.adjustPosition()

    def adjustPosition(self):
        """ adjustPosition() -> None
        Adjust the resizer

        """
        rect = self.sheet.getCellGlobalRect(self.row, self.col)
        if self.resizer:
            self.resizer.adjustPosition(rect)

    def show(self):
        """ show() -> None
        An helper function derived from setVisible

        """
        self.setVisible(True)

    def hide(self):
        """ hide() -> None
        An helper function derived from setVisible

        """
        self.setVisible(False)

    def setVisible(self,b):
        """ setVisible(b: boolean) -> None
        Show/hide the cell helpers
        """
        if self.resizer:
            self.resizer.setVisible(b)
        if not b and self.resizer:
            self.resizer.setDragging(False)

    def isInteracting(self):
        """ isInteracting() -> boolean
        Check to see if the helper is in action with the resizer

        """
        if self.resizer:
            return self.resizer.dragging
        else:
            return False



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
import sys
import os

def run(py_dirname, in_dirname, out_dirname, filename='examples_list.txt'):
    f = open(filename)
    total = 0
    total_good = 0
    for line in f:
        # print line
        arr = line.rstrip().split(None, 2)
        total += 1
        if arr[1][0] == 'Y':
            total_good += 1

            vt_file = os.path.splitext(arr[0])[0] + '.vt'
            py_file = arr[0]

            vt_name = os.path.join(in_dirname, vt_file)
            out_vt_name = os.path.join(out_dirname, vt_file)
            py_name = os.path.join(py_dirname, py_file)
            
            py_name = os.path.join(os.path.dirname(py_name),
                                   "Python",
                                   os.path.basename(py_name))
            out_py_name = os.path.join(out_dirname, py_file)
            to_make_dirs = []
            a_dirname = os.path.dirname(out_vt_name)
            while not os.path.exists(a_dirname):
                to_make_dirs.append(a_dirname)
                a_dirname = os.path.dirname(a_dirname)
            for a_dirname in reversed(to_make_dirs):
                # print "os.mkdir('%s')" % a_dirname
                os.mkdir(a_dirname)
            
            # print "os.system('cp %s %s')" % (py_name, out_py_name)
            # print "os.system('cp %s %s')" % (vt_name, out_vt_name)
            os.system('cp %s %s' % (py_name, out_py_name))
            os.system('cp %s %s' % (vt_name, out_vt_name))
    print "%d of %d" % (total_good, total)

if __name__ == '__main__':
    if len(sys.argv) < 5:
        print "Usage: python %s " % sys.argv[0] + \
            "<python_dir> <in_dir> <out_dir> [examples_list]"
        sys.exit(-1)
    run(*sys.argv[1:])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
This module describes a core theme structure for VisTrails GUI. It
specifies measurements
"""

from __future__ import division

from vistrails.core.utils.color import ColorByName
import vistrails.core.system
################################################################################

class DefaultCoreTheme(object):
    """
    This is the default core theme which contains non-qt values,
    measurements for Vistrail. Other core themes should derive from
    this class and change appropriate values
    
    """
    
    def __init__(self):
        """ DefaultTheme() -> DefaultTheme
        This is for initializing all non-Qt values
        
        """
        ######################
        #### MEASUREMENTS ####

        # Padded space of Version shape and its label
        self.VERSION_LABEL_MARGIN = (60, 35)

        # Padded space of Module shape into its label
        self.MODULE_LABEL_MARGIN = (20, 20, 20, 15)

        # Padded space of Module shape into its edit widget
        self.MODULE_EDIT_MARGIN = (8, 4, 8, 4)

        # Margin of Module shape into its ports
        self.MODULE_PORT_MARGIN = (4, 4, 4, 4)

        # Space between ports inside a module
        self.MODULE_PORT_SPACE = 4

        # The space added to the end of port shapes before it reaches the
        # margin of the module
        self.MODULE_PORT_PADDED_SPACE = 20

        # Width and Height of Port shape
        self.PORT_WIDTH = 10
        self.PORT_HEIGHT = 10

        # Width and Height of Configure button shape
        self.CONFIGURE_WIDTH = 6
        self.CONFIGURE_HEIGHT = 10

        self.BREAKPOINT_FRINGE = \
            (((0.0,0.0),(-0.5,0.25),(-0.5,0.75),(0.0,1.0)),
             ((0.0,0.0),(0.5,0.25),(0.5,0.75),(0.0,1.0)))
                                       

        # The number of control points when drawing connection curve
        self.CONNECTION_CONTROL_POINTS = 20

        # Control the size and gap for the 3 little segments when
        # draw connections between versions
        self.LINK_SEGMENT_LENGTH = 15
        self.LINK_SEGMENT_GAP = 5
        self.LINK_SEGMENT_SQUARE_LENGTH = 12

        # The size of the frame containing the PIP graphics view
        self.PIP_IN_FRAME_WIDTH = 5
        self.PIP_OUT_FRAME_WIDTH = 1

        # The size of the frame containing the PIP graphics view
        self.PIP_DEFAULT_SIZE = (128, 128)

        # The default minimum size of the graphics views
        self.BOUNDING_RECT_MINIMUM = 512

        # Default Paramter Inspector Window dimension
        self.VISUAL_DIFF_PARAMETER_WINDOW_SIZE = (348,256)

        # Default legend size (small rectangular shape)
        self.VISUAL_DIFF_LEGEND_SIZE = (16, 16)

        # Virtual Cell Label default  size
        self.VIRTUAL_CELL_LABEL_SIZE = (40, 40)

        # Query Preview Size
        self.QUERY_PREVIEW_SIZE = (256, 256)
